# Copyright © 2016 CZ.NIC, z. s. p. o.
#
# This file is part of Yangson.
#
# Yangson is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Yangson is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with Yangson.  If not, see <http://www.gnu.org/licenses/>.

"""XPath node-set"""

from typing import List, Callable, Union
from numbers import Number
from .instance import InstanceNode

# Type aliases

NodeExpr = Callable[[InstanceNode], "NodeSet"]
XPathValue = Union["NodeSet", str, float, bool]

def comparison(meth):
    def wrap(self, arg):
        if isinstance(arg, NodeSet):
            for n in arg:
                if n.is_internal(): continue
                if meth(self, str(n)):
                    return True
            return False
        return meth(self, arg)
    return wrap

class NodeSet(list):

    def union(self, ns: "NodeSet") -> "NodeSet":
        paths = set([n.json_pointer() for n in self])
        return self.__class__(self + [n for n in ns if n.json_pointer() not in paths])

    def bind(self, trans: NodeExpr) -> "NodeSet":
        res = self.__class__([])
        for n in self:
            res = res.union(trans(n))
        return res

    def __float__(self) -> float:
        return float(self[0].value)

    def __str__(self) -> str:
        return str(self[0]) if self else ""

    @comparison
    def __eq__(self, val: XPathValue) -> bool:
        for n in self:
            if n.is_internal(): continue
            if isinstance(val, str):
                if str(n) == val:
                    return True
            elif isinstance(n.value, Number):
                if float(n.value) == val:
                    return True
            elif n.value == val:
                return True
        return False

    @comparison
    def __ne__(self, val: XPathValue) -> bool:
        for n in self:
            if n.is_internal(): continue
            if isinstance(val, str):
                if str(n) != val:
                    return True
            elif isinstance(n.value, Number):
                if float(n.value) != val:
                    return True
            elif n.value != val:
                return True
        return False

    @comparison
    def __gt__(self, val: XPathValue) -> bool:
        try:
            val = float(val)
        except (ValueError, TypeError):
            return False
        for n in self:
            try:
                if float(n.value) > val:
                    return True
            except (ValueError, TypeError):
                continue
        return False

    @comparison
    def __lt__(self, val: XPathValue) -> bool:
        try:
            val = float(val)
        except (ValueError, TypeError):
            return False
        for n in self:
            try:
                if float(n.value) < val:
                    return True
            except (ValueError, TypeError):
                continue
        return False

    @comparison
    def __ge__(self, val: XPathValue) -> bool:
        try:
            val = float(val)
        except (ValueError, TypeError):
            return False
        for n in self:
            try:
                if float(n.value) >= val:
                    return True
            except (ValueError, TypeError):
                continue
        return False

    @comparison
    def __le__(self, val: XPathValue) -> bool:
        try:
            val = float(val)
        except (ValueError, TypeError):
            return False
        for n in self:
            try:
                if float(n.value) <= val:
                    return True
            except (ValueError, TypeError):
                continue
        return False



# Copyright © 2016 CZ.NIC, z. s. p. o.
#
# This file is part of Yangson.
#
# Yangson is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Yangson is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with Yangson.  If not, see <http://www.gnu.org/licenses/>.

"""
Essential data model structures and methods.

This module implements the following classes:

* ModuleData: Data related to a YANG module or submodule.
* Context: Repository of data model structures and methods.
* FeatureExprParser: Parser for if-feature expressions.

This module defines the following exceptions:

* BadPath: Invalid schema path
* BadYangLibraryData: Invalid YANG library data.
* CyclicImports: Imports of YANG modules form a cycle.
* FeaturePrerequisiteError: Pre-requisite feature isn't supported.
* InvalidFeatureExpression: Invalid if-feature expression.
* ModuleNotFound: YANG module not found.
* ModuleNotImported: YANG module is not imported.
* ModuleNotRegistered: Module is not registered in YANG library.
* MultipleImplementedRevisions: A module has multiple implemented revisions.
* UnknownPrefix: Unknown namespace prefix.
"""

from typing import Dict, List, MutableSet, Optional, Tuple
from .enumerations import ContentType
from .exceptions import YangsonException
from .parser import Parser, ParserException
from .statement import DefinitionNotFound, ModuleParser, Statement
from .typealiases import *

class ModuleData:
    """Data related to a YANG module or submodule."""

    def __init__(self, main_module: YangIdentifier):
        """Initialize the class instance."""
        self.features = set() # type: MutableSet[YangIdentifier]
        """Set of supported features."""
        self.main_module = main_module # type: ModuleId
        """Main module of the receiver."""
        self.prefix_map = {} # type: Dict[YangIdentifier, ModuleId]
        """Map of prefixes to module identifiers."""
        self.statement = None # type: Statement
        """Corresponding (sub)module statements."""
        self.submodules = set() # type: MutableSet[ModuleId]
        """Set of submodules."""

class Context:
    """Global repository of data model structures and utility methods."""

    @classmethod
    def _initialize(cls) -> None:
        """Initialize the data model structures."""
        cls.identity_bases = {} # type: Dict[QualName, MutableSet[QualName]]
        """Dictionary of identity bases."""
        cls.implement = {} # type: Dict[YangIdentifier, RevisionDate]
        """Dictionary of implemented revisions."""
        cls.module_search_path = [] # type: List[str]
        """List of directories where to look for YANG modules."""
        cls.modules = {} # type: Dict[ModuleId, ModuleData]
        """Dictionary of module data."""
        cls._module_sequence = [] # type: List[ModuleId]
        """List that defines the order of module processing."""

    @classmethod
    def _from_yang_library(cls, yang_lib: Dict[str, Any],
                           mod_path: List[str]) -> None:
        """Set the data model structures from YANG library data.

        This method requires that the class variable `schema` be
        initialized beforehand with a GroupNode instance.

        Args:
            yang_lib: Dictionary with YANG library data.
            mod_path: Value for `Context.module_search_path`.

        Raises:
            BadYangLibraryData: If YANG library data is invalid.
            FeaturePrerequisiteError: If a pre-requisite feature isn't
                supported.
            MultipleImplementedRevisions: If multiple revisions of an
                implemented module are listed in YANG library.
            ModuleNotFound: If a YANG module wasn't found in any of the
                directories specified in `mod_path`.
        """
        cls._initialize()
        cls.module_search_path = mod_path
        cls.schema._ctype = ContentType.all
        try:
            for item in yang_lib["ietf-yang-library:modules-state"]["module"]:
                name = item["name"]
                rev = item["revision"]
                mid = (name, rev)
                mdata = ModuleData(mid)
                cls.modules[mid] = mdata
                if item["conformance-type"] == "implement":
                    if name in cls.implement:
                        raise MultipleImplementedRevisions(name)
                    cls.implement[name] = rev
                mod = cls._load_module(name, rev)
                mdata.statement = mod
                if "feature" in item:
                    mdata.features.update(item["feature"])
                locpref = mod.find1("prefix", required=True).argument
                mdata.prefix_map[locpref] = mid
                if "submodule" in item:
                    for s in item["submodule"]:
                        sname = s["name"]
                        smid = (sname, s["revision"])
                        sdata = ModuleData(mid)
                        cls.modules[smid] = sdata
                        mdata.submodules.add(smid)
                        submod = cls._load_module(*smid)
                        sdata.statement = submod
                        bt = submod.find1("belongs-to", name, required=True)
                        locpref = bt.find1("prefix", required=True).argument
                        sdata.prefix_map[locpref] = mid
        except KeyError as e:
            raise BadYangLibraryData("missing " + str(e)) from None
        cls._process_imports()
        cls._check_feature_dependences()
        for mid in cls._module_sequence:
            cls.schema._new_ns = cls.namespace(mid)
            cls.schema._handle_substatements(cls.modules[mid].statement, mid)
        cls._apply_augments()
        cls.schema._post_process()
        cls.schema._make_schema_patterns()

    @classmethod
    def _load_module(cls, name: YangIdentifier,
                    rev: RevisionDate) -> Statement:
        """Read and parse a YANG module or submodule."""
        for d in cls.module_search_path:
            fn = "{}/{}".format(d, name)
            if rev: fn += "@" + rev
            fn += ".yang"
            try:
                with open(fn, encoding='utf-8') as infile:
                    res = ModuleParser(infile.read()).parse()
            except FileNotFoundError:
                continue
            return res
        raise ModuleNotFound(name, rev)

    @classmethod
    def _process_imports(cls) -> None:
        impl = set(cls.implement.items())
        deps = { mid: set() for mid in impl }
        impby = { mid: set() for mid in impl }
        for mid in cls.modules:
            mod = cls.modules[mid].statement
            for impst in mod.find_all("import"):
                impn = impst.argument
                prefix = impst.find1("prefix", required=True).argument
                revst = impst.find1("revision-date")
                if revst:
                    imid = (impn, revst.argument)
                    if imid not in cls.modules:
                        raise ModuleNotRegistered(impn, rev)
                else:                              # use last revision
                    imid = cls.last_revision(impn)
                cls.modules[mid].prefix_map[prefix] = imid
                mm = cls.modules[mid].main_module
                if mm in impl and imid in impl:
                    deps[mm].add(imid)
                    impby[imid].add(mm)
        free = [mid for mid in deps if len(deps[mid]) == 0]
        if not free: raise CyclicImports()
        while free:
            nid = free.pop()
            cls._module_sequence.append(nid)
            cls._module_sequence.extend(cls.modules[nid].submodules)
            for mid in impby[nid]:
                deps[mid].remove(nid)
                if len(deps[mid]) == 0:
                    free.append(mid)
        if [mid for mid in deps if len(deps[mid]) > 0]:
            raise CyclicImports()

    @classmethod
    def _check_feature_dependences(cls):
        """Verify feature dependences."""
        for mid in cls.modules:
            for fst in cls.modules[mid].statement.find_all("feature"):
                fn, fid = cls.resolve_pname(fst.argument, mid)
                if fn not in cls.modules[fid].features: continue
                if not cls.if_features(fst, mid):
                    raise FeaturePrerequisiteError(*fn)

    @classmethod
    def _apply_augments(cls) -> None:
        """Apply top-level augments from all implemented modules."""
        for mid in cls._module_sequence:
            mod = cls.modules[mid].statement
            for aug in mod.find_all("augment"):
                cls.schema._augment_stmt(aug, mid)

    @classmethod
    def namespace(cls, mid: ModuleId) -> YangIdentifier:
        """Return the namespace corresponding to a module or submodule.

        Args:
            mid: Module identifier.

        Raises:
            ModuleNotRegistered: If `mid` is not registered in the data model.
        """
        try:
            mdata = cls.modules[mid]
        except KeyError:
            raise ModuleNotRegistered(*mid) from None
        return mdata.main_module[0]

    @classmethod
    def last_revision(cls, mod: YangIdentifier) -> ModuleId:
        """Return the last revision of a module that's part of the data model.

        Args:
            mod: Name of a module or submodule.

        Raises:
            ModuleNotRegistered: If the module `mod` is not present in the
                data model.
        """
        revs = [mn for mn in cls.modules if mn[0] == mod]
        if not revs:
            raise ModuleNotRegistered(impn)
        return sorted(revs, key=lambda x: x[1])[-1]

    @classmethod
    def prefix2ns(cls, prefix: YangIdentifier, mid: ModuleId) -> YangIdentifier:
        """Return the namespace corresponding to a prefix.

        Args:
            prefix: Prefix associated with a module and its namespace.
            mid: Identifier of the module in which the prefix is declared.

        Raises:
            ModuleNotRegistered: If `mid` is not registered in the data model.
            UnknownPrefix: If `prefix` is not declared.
        """
        try:
            mdata = cls.modules[mid]
        except KeyError:
            raise ModuleNotRegistered(*mid) from None
        try:
            return mdata.prefix_map[prefix][0]
        except KeyError:
            raise UnknownPrefix(prefix, mid) from None

    @classmethod
    def resolve_pname(cls, pname: PrefName,
                      mid: ModuleId) -> Tuple[YangIdentifier, ModuleId]:
        """Return the name and module identifier in which the name is defined.

        Args:
            pname: Name with an optional prefix.
            mid: Identifier of the context module.

        Raises:
            ModuleNotRegistered: If `mid` is not registered in the data model.
            UnknownPrefix: If the prefix specified in `pname` is not declared.
        """
        p, s, loc = pname.partition(":")
        try:
            mdata = cls.modules[mid]
        except KeyError:
            raise ModuleNotRegistered(*mid) from None
        try:
            return (loc, mdata.prefix_map[p]) if s else (p, mdata.main_module)
        except KeyError:
            raise UnknownPrefix(p, mid) from None

    @classmethod
    def translate_pname(cls, pname: PrefName, mid: ModuleId) -> QualName:
        """Translate a prefixed name to a qualified name.

        Args:
            pname: Name with an optional prefix.
            mid: Identifier of the context module.

        Raises:
            ModuleNotRegistered: If `mid` is not registered in the data model.
            UnknownPrefix: If the prefix specified in `pname` is not declared.
        """
        loc, nid = cls.resolve_pname(pname, mid)
        return (loc, cls.namespace(nid))

    @classmethod
    def prefix(cls, imod: YangIdentifier, mid: ModuleId) -> YangIdentifier:
        """Return the prefix corresponding to an implemented module.

        Args:
            imod: Name of an implemented module.
            mid: Identifier of the context module.

        Raises:
            ModuleNotImplemented: If `imod` is not implemented.
            ModuleNotRegistered: If `mid` is not registered in YANG library.
            ModuleNotImported: If `imod` is not imported in `mid`.
        """
        try:
            did = (imod, cls.implement[imod])
        except KeyError:
            raise ModuleNotImplemented(imod) from None
        try:
            pmap = cls.modules[mid].prefix_map
        except KeyError:
            raise ModuleNotRegistered(*mid) from None
        for p in pmap:
            if pmap[p] == did:
                return p
        raise ModuleNotImported(imod, mid)

    @classmethod
    def sni2route(cls, sni: SchemaNodeId, mid: ModuleId) -> SchemaRoute:
        """Translate schema node identifier to a schema route.

        Args:
            sni: Schema node identifier (absolute or relative).
            mid: Identifier of the context module.

        Raises:
            ModuleNotRegistered: If `mid` is not registered in the data model.
            UnknownPrefix: If a prefix specified in `sni` is not declared.
        """
        nlist = sni.split("/")
        return [ cls.translate_pname(qn, mid)
                 for qn in (nlist[1:] if sni[0] == "/" else nlist) ]

    @staticmethod
    def path2route(path: SchemaPath) -> SchemaRoute:
        """Translate a schema/data path to a schema/data route.

        Args:
            path: Schema path.

        Raises:
            BadPath: Invalid path.
        """
        if path == "/" or path == "": return []
        nlist = path.split("/")
        prevns = None
        res = []
        for n in (nlist[1:] if path[0] == "/" else nlist):
            p, s, loc = n.partition(":")
            if s:
                if p == prevns: raise BadPath(path)
                res.append((loc, p))
                prevns = p
            elif prevns:
                res.append((p, prevns))
            else:
                raise BadPath(path)
        return res

    @classmethod
    def get_definition(cls, stmt: Statement,
                       mid: ModuleId) -> Tuple[Statement, ModuleId]:
        """Find the statement defining a grouping or derived type.

        Args:
            stmt: YANG "uses" or "type" statement.
            mid: Identifier of the context module.

        Returns:
            A tuple consisting of the definition statement ('grouping' or
            'typedef') and indentifier of the module where it appears.

        Raises:
            ValueError: If `stmt` is neither "uses" nor "type" statement.
            ModuleNotRegistered: If `mid` is not registered in the data model.
            UnknownPrefix: If the prefix specified in the argument of `stmt`
                is not declared.
            DefinitionNotFound: If the corresponding definition is not found.
        """
        if stmt.keyword == "uses":
            kw = "grouping"
        elif stmt.keyword == "type":
            kw = "typedef"
        else:
            raise ValueError("not a 'uses' or 'type' statement")
        loc, did = cls.resolve_pname(stmt.argument, mid)
        if did == mid:
            return (stmt.get_definition(loc, kw), mid)
        dstmt = cls.modules[did].statement.find1(kw, loc)
        if dstmt: return (dstmt, did)
        for sid in cls.modules[did].submodules:
            dstmt = cls.modules[sid].statement.find1(kw, loc)
            if dstmt: return (dstmt, sid)
        raise DefinitionNotFound(kw, stmt.argument)

    @classmethod
    def is_derived_from(cls, identity: QualName, base: QualName) -> bool:
        """Return ``True`` if `identity` is derived from `base`."""
        try:
            bases = cls.identity_bases[identity]
        except KeyError:
            return False
        if base in bases: return True
        for ib in bases:
            if cls.is_derived_from(ib, base): return True
        return False

    @classmethod
    def if_features(cls, stmt: Statement, mid: ModuleId) -> bool:
        """Evaluate ``if-feature`` substatements on a statement, if any.

        Args:
            stmt: Yang statement that is tested on if-features.
            mid: Identifier of the context module.

        Raises:
            ModuleNotRegistered: If `mid` is not registered in the data model.
            InvalidFeatureExpression: If a if-feature expression is not
                syntactically correct.
            UnknownPrefix: If a prefix specified in a feature name is not
                declared.
        """
        iffs = stmt.find_all("if-feature")
        if not iffs:
            return True
        for i in iffs:
            if not FeatureExprParser(i.argument, mid).parse():
                return False
        return True

class FeatureExprParser(Parser):
    """Parser and evaluator for if-feature expressions."""

    def __init__(self, text: str, mid: ModuleId):
        """Initialize the parser instance.

        Args:
            mid: Identifier of the context module.

        Raises:
            ModuleNotRegistered: If `mid` is not registered in the data model.
        """
        super().__init__(text)
        self.mid = mid

    def parse(self) -> bool:
        """Parse and evaluate a complete feature expression.

        Raises:
            InvalidFeatureExpression: If the if-feature expression is not
                syntactically correct.
            UnknownPrefix: If a prefix of a feature name is not declared.
        """
        self.skip_ws()
        res = self._feature_disj()
        self.skip_ws()
        if not self.at_end():
            raise InvalidFeatureExpression(self)
        return res

    def _feature_disj(self) -> bool:
        x = self._feature_conj()
        if self.test_string("or"):
            if not self.skip_ws():
                raise InvalidFeatureExpression(self)
            return self._feature_disj() or x
        return x

    def _feature_conj(self) -> bool:
        x = self._feature_term()
        if self.test_string("and"):
            if not self.skip_ws():
                raise InvalidFeatureExpression(self)
            return self._feature_conj() and x
        return x

    def _feature_term(self) -> bool:
        if self.test_string("not"):
            if not self.skip_ws():
                raise InvalidFeatureExpression(self)
            return not self._feature_atom()
        return self._feature_atom()

    def _feature_atom(self) -> bool:
        if self.peek() == "(":
            self.adv_skip_ws()
            res = self._feature_disj()
            self.char(")")
            self.skip_ws()
            return res
        n, p = self.prefixed_name()
        self.skip_ws()
        try:
            mdata = Context.modules[self.mid]
        except KeyError:
            raise ModuleNotRegistered(*self.mid) from None
        if p is None:
            fid = mdata.main_module
        else:
            try:
                fid = mdata.prefix_map[p]
            except KeyError:
                raise UnknownPrefix(p, self.mid) from None
        return n in Context.modules[fid].features

class MissingModule(YangsonException):
    """Abstract exception class – a module is missing."""

    def __init__(self, name: YangIdentifier, rev: str = ""):
        self.name = name
        self.rev = rev

    def __str__(self) -> str:
        if self.rev:
            return self.name + "@" + self.rev
        return self.name

class ModuleNotFound(MissingModule):
    """A module or submodule registered in YANG library is not found."""
    pass

class ModuleNotRegistered(MissingModule):
    """A module is not registered in YANG library."""
    pass

class ModuleNotImplemented(MissingModule):
    """A module is not implemented."""
    pass

class BadYangLibraryData(YangsonException):
    """Broken YANG library data."""

    def __init__(self, reason: str):
        self.reason = reason

    def __str__(self) -> str:
        return self.reason

class BadPath(YangsonException):
    """Invalid schema or data path."""

    def __init__(self, path: str):
        self.path = path

    def __str__(self) -> str:
        return self.path

class UnknownPrefix(YangsonException):
    """Unknown namespace prefix."""

    def __init__(self, prefix: YangIdentifier, mid: ModuleId):
        self.prefix = prefix
        self.mid = mid

    def __str__(self) -> str:
        return "prefix {} is not defined in {}".format(self.prefix, self.mid)

class ModuleNotImported(YangsonException):
    """Module is not imported."""

    def __init__(self, mod: YangIdentifier, mid: ModuleId):
        self.mod = mod
        self.mid = mid

    def __str__(self) -> str:
        return "{} not imported in {}".format(self.mod, self.mid)

class InvalidFeatureExpression(ParserException):
    """Invalid **if-feature** expression."""
    pass

class FeaturePrerequisiteError(YangsonException):
    """Pre-requisite feature is not supported."""

    def __init__(self, name: YangIdentifier, ns: YangIdentifier):
        self.name = name
        self.ns = ns

    def __str__(self) -> str:
        return "{}:{}".format(self.ns, self.name)

class MultipleImplementedRevisions(YangsonException):
    """A module has multiple implemented revisions."""

    def __init__(self, module: YangIdentifier):
        self.module = module

    def __str__(self) -> str:
        return self.module

class CyclicImports(YangsonException):
    """YANG modules are imported in a cyclic fashion."""
    pass



# Copyright © 2016 CZ.NIC, z. s. p. o.
#
# This file is part of Yangson.
#
# Yangson is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Yangson is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with Yangson.  If not, see <http://www.gnu.org/licenses/>.

"""Classes representing YANG schema nodes.

This module implements the following classes:

* SchemaNode: Abstract class for schema nodes.
* InternalNode: Abstract class for schema nodes that have children.
* GroupNode: Anonymous group of schema nodes.
* DataNode: Abstract class for data nodes.
* TerminalNode: Abstract class for schema nodes that have no children.
* ContainerNode: YANG container node.
* SequenceNode: Abstract class for schema nodes that represent a sequence.
* ListNode: YANG list node.
* ChoiceNode: YANG choice node.
* CaseNode: YANG case node.
* RpcActionNode: YANG rpc or action node.
* InputNode: YANG input node.
* OutputNode: YANG output node.
* NotificationNode: YANG notification node.
* LeafNode: YANG leaf node.
* LeafListNode: YANG leaf-list node.
* AnydataNode: YANG anydata or anyxml node.

This module defines the following exceptions:

* SchemaNodeException: Abstract exception class for schema node errors.
* NonexistentSchemaNode: A schema node doesn't exist.
* BadSchemaNodType: A schema node is of a wrong type.
* BadLeafrefPath: A leafref path is incorrect.
* ValidationError: Abstract exeption class for instance validation errors.
* SchemaError: An instance violates a schema constraint.
* SemanticError: An instance violates a semantic rule.
"""

from typing import Dict, List, MutableSet, Optional, Set, Tuple, Union
from .exceptions import YangsonException
from .context import Context
from .datatype import (DataType, LeafrefType, LinkType,
                       RawScalar, IdentityrefType)
from .enumerations import Axis, ContentType, DefaultDeny
from .instvalue import ArrayValue, EntryValue, ObjectValue, Value
from .schpattern import *
from .statement import Statement, WrongArgument
from .typealiases import *
from .xpathparser import XPathParser

class SchemaNode:
    """Abstract class for all schema nodes."""

    def __init__(self):
        """Initialize the class instance."""
        self.name = None # type: Optional[YangIdentifier]
        """Name of the schema node."""
        self.ns = None # type: Optional[YangIdentifier]
        """Namespace of the schema node."""
        self.parent = None # type: Optional["InternalNode"]
        """Parent schema node."""
        self.must = [] # type: List[Tuple["Expr", Optional[str]]]
        """List of "must" expressions attached to the schema node."""
        self.when = None # type: Optional["Expr"]
        """Optional "when" expression that makes the schema node conditional."""
        self._ctype = None
        """Content type of the schema node."""

    @property
    def qual_name(self) -> QualName:
        """Qualified name of the receiver."""
        return (self.name, self.ns)

    @property
    def config(self) -> bool:
        """Does the receiver (also) represent configuration?"""
        return self.content_type().value & ContentType.config.value != 0

    @property
    def mandatory(self) -> bool:
        """Is the receiver a mandatory node?"""
        return False

    def content_type(self) -> ContentType:
        """Return receiver's content type."""
        return self._ctype if self._ctype else self.parent.content_type()

    def data_parent(self) -> Optional["InternalNode"]:
        """Return the closest ancestor data node."""
        parent = self.parent
        while parent:
            if isinstance(parent, DataNode):
                return parent
            parent = parent.parent

    def iname(self) -> InstanceName:
        """Return the instance name corresponding to the receiver."""
        dp = self.data_parent()
        return (self.name if dp and self.ns == dp.ns
                else self.ns + ":" + self.name)

    def data_path(self) -> DataPath:
        """Return the receiver's data path."""
        dp = self.data_parent()
        return (dp.data_path() if dp else "") + "/" + self.iname()

    def state_roots(self) -> List[DataPath]:
        """Return a list of data paths to descendant state data roots."""
        return [r.data_path() for r in self._state_roots()]

    def validate(self, inst: "InstanceNode", ctype: ContentType) -> None:
        """Validate instance against the receiver.

        Args:
            inst: Instance node to be validated.
            ctype: Content type of the instance.

        Returns:
            ``None`` if validation succeeds.

        Raises:
            SchemaError: if `inst` violates the schema.
            SemanticError: If a "must" expression evaluates to ``False``.
        """
        pass

    def from_raw(self, rval: RawValue) -> Value:
        """Return instance value transformed from a raw value using receiver.

        Args:
            rval: Raw value.

        Raises:
            NonexistentSchemaNode: If a member inside `rval` is not defined
                in the schema.
            YangTypeError: If a scalar value inside `rval` is of incorrect type.
        """
        raise NotImplementedError

    def _iname2qname(self, iname: InstanceName) -> QualName:
        """Translate instance name to qualified name in the receiver's context.
        """
        p, s, loc = iname.partition(":")
        return (loc, p) if s else (p, self.ns)

    def _flatten(self) -> List["SchemaNode"]:
        return [self]

    def _handle_substatements(self, stmt: Statement, mid: ModuleId) -> None:
        """Dispatch actions for substatements of `stmt`."""
        for s in stmt.substatements:
            if s.prefix:
                key = Context.modules[mid].prefix_map[s.prefix][0] + ":" + s.keyword
            else:
                key = s.keyword
            mname = SchemaNode._stmt_callback.get(key, "_noop")
            method = getattr(self, mname)
            method(s, mid)

    def _follow_leafref(self, xpath: "Expr") -> Optional["DataNode"]:
        """Return the data node referred to by a leafref path.

        Args:
            xpath: XPath expression compiled from a leafref path.
        """
        if isinstance(xpath, LocationPath):
            lft = self._follow_leafref(xpath.left)
            if lft is None: return None
            return lft._follow_leafref(xpath.right)
        elif isinstance(xpath, Step):
            if xpath.axis == Axis.parent:
                return self.data_parent()
            elif xpath.axis == Axis.child:
                if isinstance(self, InternalNode) and xpath.qname:
                    return self.get_data_child(*xpath.qname)
        elif isinstance(xpath, Root):
            return Context.schema
        return None

    def _noop(self, stmt: Statement, mid: ModuleId) -> None:
        pass

    def _config_stmt(self, stmt: Statement, mid: ModuleId) -> None:
        if stmt.argument == "true" and self.parent.config:
            self._ctype = ContentType.all
        elif stmt.argument == "false":
            self._ctype = ContentType.nonconfig

    def _must_stmt(self, stmt: Statement, mid: ModuleId) -> None:
        xpp = XPathParser(stmt.argument, mid)
        mex = xpp.parse()
        if not xpp.at_end():
            raise WrongArgument(stmt)
        ems = stmt.find1("error-message")
        self.must.append((mex, ems.argument if ems else None))

    def _when_stmt(self, stmt: Statement, mid: ModuleId) -> None:
        xpp = XPathParser(stmt.argument, mid)
        wex = xpp.parse()
        if not xpp.at_end():
            raise WrongArgument(stmt)
        self.when = wex

    def _mandatory_stmt(self, stmt, mid: ModuleId) -> None:
        if stmt.argument == "true":
            self._mandatory = True
        elif stmt.argument == "false":
            self._mandatory = False

    def _post_process(self) -> None:
        pass

    def _is_identityref(self) -> bool:
        return False

    def _default_nodes(self, inst: "InstanceNode") -> List["InstanceNode"]:
        return []

    def _tree_line(self) -> str:
        """Return the receiver's contribution to tree diagram."""
        return self._tree_line_prefix() + " " + self.iname()

    def _tree_line_prefix(self) -> str:
        return "+--"

    _stmt_callback = {
        "action": "_rpc_action_stmt",
        "anydata": "_anydata_stmt",
        "anyxml": "_anydata_stmt",
        "case": "_case_stmt",
        "choice": "_choice_stmt",
        "config": "_config_stmt",
        "container": "_container_stmt",
        "default": "_default_stmt",
        "identity": "_identity_stmt",
        "ietf-netconf-acm:default-deny-all": "_nacm_default_deny_stmt",
        "ietf-netconf-acm:default-deny-write": "_nacm_default_deny_stmt",
        "input": "_input_stmt",
        "key": "_key_stmt",
        "leaf": "_leaf_stmt",
        "leaf-list": "_leaf_list_stmt",
        "list": "_list_stmt",
        "mandatory": "_mandatory_stmt",
        "max-elements": "_max_elements_stmt",
        "min-elements": "_min_elements_stmt",
        "must": "_must_stmt",
        "notification": "_notification_stmt",
        "output": "_output_stmt",
        "ordered-by": "_ordered_by_stmt",
        "presence": "_presence_stmt",
        "rpc": "_rpc_action_stmt",
        "unique": "_unique_stmt",
        "uses": "_uses_stmt",
        "when": "_when_stmt",
        }
    """Map of statement keywords to callback methods."""


class InternalNode(SchemaNode):
    """Abstract class for schema nodes that have children."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self.children = [] # type: List[SchemaNode]
        self._new_ns = None # type: Optional[ModuleId]
        self._mandatory_children = set() # type: MutableSet[SchemaNode]

    @property
    def mandatory(self) -> bool:
        """Is the receiver a mandatory node?"""
        return len(self._mandatory_children) > 0

    def get_child(self, name: YangIdentifier,
                  ns: YangIdentifier = None) -> Optional[SchemaNode]:
        """Return receiver's schema child.

        Args:
            name: Child's name.
            ns: Child's namespace (= `self.ns` if absent).
        """
        ns = ns if ns else self.ns
        todo = []
        for child in self.children:
            if child.name is None:
                todo.append(child)
            elif child.name == name and child.ns == ns:
                return child
        for c in todo:
            return c.get_child(name, ns)

    def get_schema_descendant(
            self, route: SchemaRoute) -> Optional[SchemaNode]:
        """Return descendant schema node or ``None`` if not found.

        Args:
            route: Schema route to the descendant node
                   (relative to the receiver).
        """
        node = self
        for p in route:
            node = node.get_child(*p)
            if node is None: return None
        return node

    def get_data_child(self, name: YangIdentifier,
                       ns: YangIdentifier = None) -> Optional["DataNode"]:
        """Return data node directly under the receiver."""
        ns = ns if ns else self.ns
        todo = []
        for child in self.children:
            if child.name == name and child.ns == ns:
                if isinstance(child, DataNode): return child
                todo.insert(0, child)
            elif not isinstance(child, DataNode):
                todo.append(child)
        for c in todo:
            res = c.get_data_child(name, ns)
            if res: return res

    def filter_children(self, ctype: ContentType = None) -> List[SchemaNode]:
        """Return receiver's children based on content type.

        Args:
            ctype: Content type.
        """
        if ctype is None:
            ctype = self.content_type()
        return [c for c in self.children if
                not isinstance(c, (RpcActionNode, NotificationNode)) and
                c.content_type().value & ctype.value != 0]

    def data_children(self) -> List["DataNode"]:
        """Return the set of all data nodes directly under the receiver."""
        res = []
        for child in self.children:
            if isinstance(child, DataNode):
                res.append(child)
            else:
                res.extend(child.data_children())
        return res

    def validate(self, inst: "InstanceNode", ctype: ContentType) -> None:
        """Extend the superclass method."""
        if not isinstance(inst.value, ObjectValue):
            raise SchemaError(inst, "non-object value")
        self._check_schema_pattern(inst, ctype)
        super().validate(inst, ctype)
        for m in inst.value:
            inst.member(m).validate(ctype)

    def from_raw(self, rval: RawObject) -> ObjectValue:
        """Override the superclass method."""
        res = ObjectValue()
        for qn in rval:
            cn = self._iname2qname(qn)
            ch = self.get_data_child(*cn)
            if ch is None:
                raise NonexistentSchemaNode(self, *cn)
            res[ch.iname()] = ch.from_raw(rval[qn])
        return res

    def _add_child(self, node: SchemaNode) -> None:
        node.parent = self
        self.children.append(node)

    def _child_inst_names(self) -> Set[InstanceName]:
        """Return the set of instance names under the receiver."""
        return frozenset([c.iname() for c in self.data_children()])

    def _check_schema_pattern(self, inst: "InstanceNode",
                             ctype: ContentType) -> None:
        """Match instance value against receiver's schema pattern.

        Args:
            inst: Instance node to be chancked.
            ctype: Content type of the instance.

        Raises:
            SchemaError: if `inst` doesn't match the schema pattern.
        """
        p = self.schema_pattern
        p._eval_when(inst)
        for m in inst.value:
            p = p.deriv(m, ctype)
            if isinstance(p, NotAllowed):
                raise SchemaError(inst, "not allowed: member {}{}".format(
                    m, ("" if ctype == ContentType.all else
                        " (" + ctype.name + ")")))
        if not p.nullable(ctype):
            raise SchemaError(inst, "missing: " + str(p))

    def _make_schema_patterns(self) -> None:
        """Build schema pattern for the receiver and its data descendants."""
        self.schema_pattern = self._schema_pattern()
        for dc in self.data_children():
            if isinstance(dc, InternalNode):
                dc._make_schema_patterns()

    def _schema_pattern(self) -> SchemaPattern:
        todo = [c for c in self.children
                if not isinstance(c, (RpcActionNode, NotificationNode))]
        if not todo: return Empty()
        prev = todo[0]._pattern_entry()
        for c in todo[1:]:
            prev = Pair(c._pattern_entry(), prev)
        return ConditionalPattern(prev, self.when) if self.when else prev

    def _post_process(self) -> None:
        super()._post_process()
        for c in self.children:
            c._post_process()

    def _add_mandatory_child(self, node: SchemaNode) -> None:
        """Add `node` to the set of mandatory children."""
        self._mandatory_children.add(node)

    def _add_defaults(self, inst: "InstanceNode", ctype: ContentType,
                      lazy: bool = False) -> "InstanceNode":
        for c in self.filter_children(ctype):
            if isinstance(c, DataNode):
                inst = c._default_instance(inst, ctype, lazy)
            elif not isinstance(c, (RpcActionNode, NotificationNode)):
                inst = c._add_defaults(inst, ctype)
        return inst

    def _state_roots(self) -> List[SchemaNode]:
        if self.content_type() == ContentType.nonconfig:
            return [self]
        res = []
        for c in self.data_children():
            res.extend(c._state_roots())
        return res

    def _handle_child(
            self, node: SchemaNode, stmt: Statement, mid: ModuleId) -> None:
        """Add child node to the receiver and handle substatements."""
        if not Context.if_features(stmt, mid): return
        node.name = stmt.argument
        node.ns = self._new_ns if self._new_ns else self.ns
        self._add_child(node)
        node._handle_substatements(stmt, mid)

    def _augment_stmt(self, stmt: Statement, mid: ModuleId) -> None:
        """Handle **augment** statement."""
        if not Context.if_features(stmt, mid): return
        path = Context.sni2route(stmt.argument, mid)
        target = self.get_schema_descendant(path)
        if stmt.find1("when"):
            gr = GroupNode()
            target._add_child(gr)
            target = gr
        myns = Context.namespace(mid)
        target._new_ns = None if target.ns == myns else myns
        target._handle_substatements(stmt, mid)

    def _refine_stmt(self, stmt: Statement, mid: ModuleId) -> None:
        """Handle **refine** statement."""
        target = self.get_schema_descendant(
            Context.sni2route(stmt.argument, mid))
        if not Context.if_features(stmt, mid):
            target.parent.children.remove(target)
        else:
            target._handle_substatements(stmt, mid)

    def _uses_stmt(self, stmt: Statement, mid: ModuleId) -> None:
        """Handle uses statement."""
        if not Context.if_features(stmt, mid): return
        grp, gid = Context.get_definition(stmt, mid)
        if stmt.find1("when"):
            sn = GroupNode()
            self._add_child(sn)
        else:
            sn = self
        sn._handle_substatements(grp, gid)
        for augst in stmt.find_all("augment"):
            sn._augment_stmt(augst, mid)
        for refst in stmt.find_all("refine"):
            sn._refine_stmt(refst, mid)

    def _container_stmt(self, stmt: Statement, mid: ModuleId) -> None:
        """Handle container statement."""
        self._handle_child(ContainerNode(), stmt, mid)

    def _identity_stmt(self, stmt: Statement, mid: ModuleId) -> None:
        """Handle identity statement."""
        if not Context.if_features(stmt, mid): return
        bases = stmt.find_all("base")
        Context.identity_bases[
            (stmt.argument, Context.namespace(mid))] = set(
                [Context.translate_pname(ist.argument, mid) for ist in bases])

    def _list_stmt(self, stmt: Statement, mid: ModuleId) -> None:
        """Handle list statement."""
        self._handle_child(ListNode(), stmt, mid)

    def _choice_stmt(self, stmt: Statement, mid: ModuleId) -> None:
        """Handle choice statement."""
        self._handle_child(ChoiceNode(), stmt, mid)

    def _case_stmt(self, stmt: Statement, mid: ModuleId) -> None:
        """Handle case statement."""
        self._handle_child(CaseNode(), stmt, mid)

    def _leaf_stmt(self, stmt: Statement, mid: ModuleId) -> None:
        """Handle leaf statement."""
        node = LeafNode()
        node.type = DataType._resolve_type(
            stmt.find1("type", required=True), mid)
        self._handle_child(node, stmt, mid)

    def _leaf_list_stmt(self, stmt: Statement, mid: ModuleId) -> None:
        """Handle leaf-list statement."""
        node = LeafListNode()
        node.type = DataType._resolve_type(
            stmt.find1("type", required=True), mid)
        self._handle_child(node, stmt, mid)

    def _rpc_action_stmt(self, stmt: Statement, mid: ModuleId) -> None:
        """Handle rpc or action statement."""
        self._handle_child(RpcActionNode(), stmt, mid)

    def _notification_stmt(self, stmt: Statement, mid: ModuleId) -> None:
        """Handle notification statement."""
        self._handle_child(NotificationNode(), stmt, mid)

    def _anydata_stmt(self, stmt: Statement, mid: ModuleId) -> None:
        """Handle anydata statement."""
        self._handle_child(AnydataNode(), stmt, mid)

    def _ascii_tree(self, indent: str) -> str:
        """Return the receiver's subtree as ASCII art."""
        if not self.children: return ""
        cs = []
        for c in self.children:
            cs.extend(c._flatten())
        cs.sort(key=lambda x: x.qual_name)
        res = ""
        for c in cs[:-1]:
            res += (indent + c._tree_line() + "\n" +
                    c._ascii_tree(indent + "|  "))
        return (res + indent + cs[-1]._tree_line() + "\n" +
                cs[-1]._ascii_tree(indent + "   "))

class GroupNode(InternalNode):
    """Anonymous group of schema nodes."""

    def state_roots(self) -> List[DataPath]:
        """Override superclass method."""
        return []

    def _state_roots(self) -> List[SchemaNode]:
        return []

    def _handle_child(self, node: SchemaNode, stmt: Statement,
                     mid: ModuleId) -> None:
        if not isinstance(self.parent, ChoiceNode) or isinstance(node, CaseNode):
            super()._handle_child(node, stmt, mid)
        else:
            cn = CaseNode()
            cn.name = stmt.argument
            cn.ns = self._new_ns if self._new_ns else self.ns
            self._add_child(cn)
            cn._handle_child(node, stmt, mid)

    def _pattern_entry(self) -> SchemaPattern:
        return super()._schema_pattern()

    def _flatten(self) -> List[SchemaNode]:
        res = []
        for c in self.children:
            res.extend(c._flatten())
        return res

class DataNode(SchemaNode):
    """Abstract superclass for all data nodes."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self.default_deny = DefaultDeny.none # type: "DefaultDeny"

    def validate(self, inst: "InstanceNode", ctype: ContentType) -> None:
        """Extend the superclass method."""
        self._check_must(inst)
        super().validate(inst, ctype)

    def _default_instance(self, pnode: "InstanceNode", ctype: ContentType,
                          lazy: bool = False) -> "InstanceNode":
        iname = self.iname()
        if iname in pnode.value: return pnode
        nm = pnode.put_member(iname, (None,))
        if not self.when or self.when.evaluate(nm):
            wd = self._default_value(nm, ctype, lazy)
            if wd.value is not None:
                return wd.up()
        return pnode

    def _check_must(self, inst: "InstanceNode") -> None:
        """Check that all receiver's "must" constraints for the instance.

        Args:
            inst: Instance node to be checked.

        Raises:
            SemanticError: If a "must" expression evaluates to ``False``.
        """
        for mex in self.must:
            if not mex[0].evaluate(inst):
                msg = "'must' expression is false" if mex[1] is None else mex[1]
                raise SemanticError(inst, msg)

    def _pattern_entry(self) -> SchemaPattern:
        m = Member(self.iname(), self.content_type(), self.when)
        return m if self.mandatory else SchemaPattern.optional(m)

    def _tree_line_prefix(self) -> str:
        return super()._tree_line_prefix() + (
            "ro" if self.content_type() == ContentType.nonconfig else "rw")

    def _nacm_default_deny_stmt(self, stmt: Statement, mid: ModuleId) -> None:
        """Set NACM default access."""
        if stmt.keyword == "default-deny-all":
            self.default_deny = DefaultDeny.all
        elif stmt.keyword == "default-deny-write":
            self.default_deny = DefaultDeny.write

class TerminalNode(SchemaNode):
    """Abstract superclass for terminal nodes in the schema tree."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self.type = None # type: DataType
        self._default = None # type: Optional[Value]

    def content_type(self) -> ContentType:
        """Override superclass method."""
        if self._ctype:
            return self._ctype
        return (ContentType.config if self.parent.config else
                ContentType.nonconfig)

    def validate(self, inst: "InstanceNode", ctype: ContentType) -> None:
        """Extend the superclass method."""
        self._check_type(inst)
        super().validate(inst, ctype)

    def from_raw(self, rval: RawScalar) -> ScalarValue:
        """Override the superclass method."""
        return self.type.from_raw(rval)

    def _default_value(self, inst: "InstanceNode", ctype: ContentType,
                       lazy: bool) -> "InstanceNode":
        inst.value = self.default
        return inst

    def _check_type(self, inst: "InstanceNode"):
        """Check whether receiver's type matches the instance value.

        Args:
            inst: Instance node to be checked.

        Raises:
            SchemaError: If the instance value doesn't match the type.
            SemanticError: If the instance violates referential integrity.
        """
        if not self.type.contains(inst.value):
            raise SchemaError(inst, "invalid type: " + repr(inst.value))
        if (isinstance(self.type, LinkType) and self.type.require_instance and
            not inst._deref()):
            raise SemanticError(inst, "required instance missing")

    def _post_process(self) -> None:
        super()._post_process()
        if isinstance(self.type, LeafrefType):
            ref = self._follow_leafref(self.type.path)
            if ref is None:
                raise BadLeafrefPath(self)
            self.type.ref_type = ref.type

    def _is_identityref(self) -> bool:
        return isinstance(self.type, IdentityrefType)

    def _default_nodes(self, inst: "InstanceNode") -> List["InstanceNode"]:
        di = self._default_instance(inst, ContentType.all)
        return [] if di is None else [self]
        return inst.put_member(self.iname(), dflt)._node_set()

    def _ascii_tree(self, indent: str) -> str:
        return ""

    def _state_roots(self) -> List[SchemaNode]:
        return [] if self.content_type() == ContentType.config else [self]

class ContainerNode(DataNode, InternalNode):
    """Container node."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self.presence = False # type: bool

    @property
    def mandatory(self) -> bool:
        """Is the receiver a mandatory node?"""
        return not self.presence and super().mandatory

    def _add_mandatory_child(self, node: SchemaNode):
        if not (self.presence or self.mandatory):
            self.parent._add_mandatory_child(self)
        super()._add_mandatory_child(node)

    def _default_instance(self, pnode: "InstanceNode", ctype: ContentType,
                          lazy: bool = False) -> "InstanceNode":
        if self.presence:
            return pnode
        return super()._default_instance(pnode, ctype, lazy)

    def _default_value(self, inst: "InstanceNode", ctype: ContentType,
                       lazy: bool) -> Optional["InstanceNode"]:
        inst.value = ObjectValue()
        return inst if lazy else self._add_defaults(inst, ctype)

    def _default_nodes(self, inst: "InstanceNode") -> List["InstanceNode"]:
        if self.presence: return []
        res = inst.put_member(self.iname(), ObjectValue())
        if self.when is None or self.when.evaluate(res):
            return [res]
        return []

    def _presence_stmt(self, stmt: Statement, mid: ModuleId) -> None:
        self.presence = True

    def _tree_line(self) -> str:
        """Return the receiver's contribution to tree diagram."""
        return super()._tree_line() + ("!" if self.presence else "")

class SequenceNode(DataNode):
    """Abstract class for data nodes that represent a sequence."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self.min_elements = 0 # type: int
        self.max_elements = None # type: Optional[int]
        self.user_ordered = False # type: bool

    @property
    def mandatory(self) -> bool:
        """Is the receiver a mandatory node?"""
        return self.min_elements > 0

    def validate(self, inst: "InstanceNode", ctype: ContentType) -> None:
        """Extend the superclass method."""
        if isinstance(inst, ArrayEntry):
            super().validate(inst, ctype)
        elif isinstance(inst.value, ArrayValue):
            self._check_unique(inst)
            self._check_cardinality(inst)
            try:
                e = inst.entry(0)
                while True:
                    super().validate(e, ctype)
                    e = e.next()
            except NonexistentInstance:
                pass
        else:
            raise SchemaError(inst, "non-array value")

    def _check_cardinality(self, inst: "InstanceNode") -> None:
        """Check that the instance satisfies cardinality constraints.

        Args:
            inst: Instance node to be checked.

        Raises:
            SchemaError: It the cardinality of `inst` isn't correct.
        """
        if len(inst.value) < self.min_elements:
            raise SemanticError(inst,
                              "number of entries < min-elements ({})".format(
                                  self.min_elements))
        if (self.max_elements is not None and
            len(inst.value) > self.max_elements):
            raise SemanticError(inst,
                              "number of entries > max-elements ({})".format(
                                  self.max_elements))

    def _post_process(self) -> None:
        super()._post_process()
        if self.min_elements > 0:
            self.parent._add_mandatory_child(self)

    def _min_elements_stmt(self, stmt: Statement, mid: ModuleId) -> None:
        self.min_elements = int(stmt.argument)

    def _max_elements_stmt(self, stmt: Statement, mid: ModuleId) -> None:
        arg = stmt.argument
        if arg == "unbounded":
            self.max_elements = None
        else:
            self.max_elements = int(arg)

    def _ordered_by_stmt(self, stmt: Statement, mid: ModuleId) -> None:
        self.user_ordered = stmt.argument == "user"

    def _tree_line(self) -> str:
        """Extend the superclass method."""
        return super()._tree_line() + "*"

    def from_raw(self, rval: RawList) -> ArrayValue:
        """Override the superclass method."""
        res = ArrayValue()
        for en in rval:
            res.append(self.entry_from_raw(en))
        return res

    def entry_from_raw(self, rval: RawEntry) -> EntryValue:
        """Transform a raw (leaf-)list entry into the cooked form.

        Args:
            rval: raw entry (scalar or object)

        Raises:
            NonexistentSchemaNode: If a member inside `rval` is not defined
                in the schema.
            YangTypeError: If a scalar value inside `rval` is of incorrect type.
        """
        return super().from_raw(rval)

class ListNode(SequenceNode, InternalNode):
    """List node."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self.keys = [] # type: List[QualName]
        self._key_members = []
        self.unique = [] # type: List[List[SchemaRoute]]

    def _check_unique(self, inst: "InstanceNode") -> None:
        """Check uniqueness of keys and "unique" properties."""
        ukeys = set()
        for i in range(len(inst.value)):
            en = inst.value[i]
            try:
                kval = tuple([en[k] for k in self._key_members])
            except KeyError as e:
                raise SchemaError(
                    inst.entry(i),
                    "missing list key '{}'".format(e.args[0])) from None
            if kval in ukeys:
                raise SchemaError(inst, "non-unique list key: " + repr(kval))
            ukeys.add(kval)
        for u in self.unique:
            uvals = set()
            try:
                en = inst.entry(0)
                while True:
                    den = en.add_defaults()
                    uval = tuple([den._peek_schema_route(sr) for sr in u])
                    if None not in uval:
                        if uval in uvals:
                            raise SemanticError(
                                inst, "unique constraint violated")
                        else:
                            uvals.add(uval)
                    en = en.next()
            except NonexistentInstance:
                continue

    def _default_instance(self, pnode: "InstanceNode", ctype: ContentType,
                          lazy: bool = False) -> "InstanceNode":
        return pnode

    def _post_process(self) -> None:
        super()._post_process()
        for k in self.keys:
            kn = self.get_data_child(*k)
            self._key_members.append(kn.iname())
            if not kn._mandatory:
                kn._mandatory = True
                self._mandatory_children.add(kn)

    def _key_stmt(self, stmt: Statement, mid: ModuleId) -> None:
        self.keys = []
        for k in stmt.argument.split():
            self.keys.append(Context.translate_pname(k, mid) if ":" in k
                             else (k, self.ns))

    def _unique_stmt(self, stmt: Statement, mid: ModuleId) -> None:
        self.unique.append(
            [Context.sni2route(sid, mid) for sid in stmt.argument.split()])

    def _tree_line(self) -> str:
        """Return the receiver's contribution to tree diagram."""
        keys = (" [" + " ".join([ k[0] for k in self.keys ]) + "]"
                if self.keys else "")
        return super()._tree_line() + keys

class ChoiceNode(InternalNode):
    """Choice node."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self.default_case = None # type: QualName
        self._mandatory = False # type: bool

    @property
    def mandatory(self) -> bool:
        """Is the receiver a mandatory node?"""
        return self._mandatory

    def _add_defaults(self, inst: "InstanceNode",
                      ctype: ContentType) -> "InstanceNode":
        if self.when and not self.when.evaluate(inst):
            return inst
        ac = self._active_case(inst.value)
        if ac:
            return ac._add_defaults(inst, ctype)
        elif self.default_case:
            n = dc = self.get_child(*self.default_case)
            while n is not self:
                if n.when and not n.when.evaluate(inst):
                    return inst
                n = n.parent
            return dc._add_defaults(inst, ctype)
        else:
            return inst

    def _active_case(self, value: ObjectValue) -> Optional["CaseNode"]:
        """Return receiver's case that's active in an instance node value."""
        for c in self.children:
            for cc in c.data_children():
                if cc.iname() in value:
                    return c

    def _pattern_entry(self) -> SchemaPattern:
        if not self.children:
            return Empty()
        prev = self.children[0]._schema_pattern()
        for c in self.children[1:]:
            prev = ChoicePattern(c._schema_pattern(), prev, self.name)
        prev.ctype = self.content_type()
        if not self.mandatory:
            prev = SchemaPattern.optional(prev)
        return ConditionalPattern(prev, self.when) if self.when else prev

    def _post_process(self) -> None:
        super()._post_process()
        if self._mandatory:
            self.parent._add_mandatory_child(self)

    def _config_stmt(self, stmt: Statement, mid: ModuleId) -> None:
        if stmt.argument == "false":
            self._ctype = ContentType.nonconfig

    def _default_nodes(self, inst: "InstanceNode") -> List["InstanceNode"]:
        res = []
        if self.default_case is None: return res
        for cn in self.get_child(*self.default_case).children:
            res.extend(cn._default_nodes(inst))
        return res

    def _tree_line_prefix(self) -> str:
        return super()._tree_line_prefix() + (
            "ro" if self.content_type() == ContentType.nonconfig else "rw")

    def _handle_child(self, node: SchemaNode, stmt: Statement,
                     mid: ModuleId) -> None:
        if isinstance(node, CaseNode):
            super()._handle_child(node, stmt, mid)
        else:
            cn = CaseNode()
            cn.name = stmt.argument
            cn.ns = self._new_ns if self._new_ns else self.ns
            self._add_child(cn)
            cn._handle_child(node, stmt, mid)

    def _default_stmt(self, stmt: Statement, mid: ModuleId) -> None:
        self.default_case = Context.translate_pname(stmt.argument, mid)

    def _tree_line(self) -> str:
        """Return the receiver's contribution to tree diagram."""
        return "{} ({}){}".format(
            self._tree_line_prefix(), self.iname(),
            "" if self._mandatory else "?")

class CaseNode(InternalNode):
    """Case node."""

    def _pattern_entry(self) -> SchemaPattern:
        return super()._schema_pattern()

    def _tree_line(self) -> str:
        """Return the receiver's contribution to tree diagram."""
        return "{}:({})".format(
            self._tree_line_prefix(), self.iname())

class LeafNode(DataNode, TerminalNode):
    """Leaf node."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self._mandatory = False # type: bool

    @property
    def mandatory(self) -> bool:
        """Is the receiver a mandatory node?"""
        return self._mandatory

    @property
    def default(self) -> Optional[ScalarValue]:
        """Default value of the receiver, if any."""
        if self.mandatory: return None
        if self._default is not None: return self._default
        return self.type.default

    def _post_process(self) -> None:
        super()._post_process()
        if self._mandatory:
            self.parent._add_mandatory_child(self)

    def _tree_line(self) -> str:
        return super()._tree_line() + ("" if self._mandatory else "?")

    def _default_stmt(self, stmt: Statement, mid: ModuleId) -> None:
        self._default = self.type.from_yang(stmt.argument, mid)

class LeafListNode(SequenceNode, TerminalNode):
    """Leaf-list node."""

    @property
    def default(self) -> Optional[ScalarValue]:
        """Default value of the receiver, if any."""
        if self.mandatory: return None
        if self._default is not None: return self._default
        return (None if self.type.default is None
                else ArrayValue([self.type.default]))

    def _check_unique(self, inst: "InstanceNode") -> None:
        if (self.content_type() == ContentType.config and
            len(set(inst.value)) < len(inst.value)):
            raise SemanticError(inst, "non-unique leaf-list values")

    def _default_stmt(self, stmt: Statement, mid: ModuleId) -> None:
        val = self.type.parse_value(stmt.argument)
        if self._default is None:
            self._default = ArrayValue([val])
        else:
            self._default.append(val)

class AnydataNode(DataNode):
    """Anydata or anyxml node."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self._mandatory = False # type: bool

    def content_type(self) -> ContentType:
        """Override superclass method."""
        return TerminalNode.content_type(self)

    @property
    def mandatory(self) -> bool:
        """Is the receiver a mandatory node?"""
        return self._mandatory

    def from_raw(self, rval: RawValue) -> Value:
        """Override the superclass method."""
        def convert(val):
            if isinstance(val, list):
                res = ArrayValue([convert(x) for x in val])
                res.stamp()
            elif isinstance(val, dict):
                res = ObjectValue({ x:convert(val[x]) for x in val })
                res.stamp()
            else:
                res = val
            return res
        return convert(rval)

    def _default_instance(self, pnode: "InstanceNode", ctype: ContentType,
                          lazy: bool = False) -> "InstanceNode":
        return pnode

    def _tree_line(self) -> str:
        return super()._tree_line() + ("" if self._mandatory else "?")

    def _ascii_tree(self, indent: str) -> str:
        return ""

    def _post_process(self) -> None:
        if self._mandatory:
            self.parent._add_mandatory_child(self)

class RpcActionNode(GroupNode):
    """RPC or action node."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self._ctype = ContentType.nonconfig

    def _handle_substatements(self, stmt: Statement, mid: ModuleId) -> None:
        self._add_child(InputNode(self.ns))
        self._add_child(OutputNode(self.ns))
        super()._handle_substatements(stmt, mid)

    def _flatten(self) -> List[SchemaNode]:
        return [self]

    def _tree_line_prefix(self) -> str:
        return super()._tree_line_prefix() + "-x"

    def _input_stmt(self, stmt: Statement, mid: ModuleId) -> None:
        """Handle RPC or action input statement."""
        self.get_child("input")._handle_substatements(stmt, mid)

    def _output_stmt(self, stmt: Statement, mid: ModuleId) -> None:
        """Handle RPC or action output statement."""
        self.get_child("output")._handle_substatements(stmt, mid)

class InputNode(GroupNode):
    """RPC or action input node."""

    def __init__(self, ns):
        """Initialize the class instance."""
        super().__init__()
        self._config = False
        self.name = "input"
        self.ns = ns

    def _flatten(self) -> List[SchemaNode]:
        return [self]

    def _tree_line_prefix(self) -> str:
        return super()._tree_line_prefix() + "ro"

class OutputNode(GroupNode):
    """RPC or action output node."""

    def __init__(self, ns):
        """Initialize the class instance."""
        super().__init__()
        self._config = False
        self.name = "output"
        self.ns = ns

    def _flatten(self) -> List[SchemaNode]:
        return [self]

    def _tree_line_prefix(self) -> str:
        return super()._tree_line_prefix() + "ro"

class NotificationNode(GroupNode):
    """Notification node."""

    def __init__(self):
        """Initialize the class instance."""
        super().__init__()
        self._ctype = ContentType.nonconfig

    def _flatten(self) -> List[SchemaNode]:
        return [self]

    def _tree_line_prefix(self) -> str:
        return super()._tree_line_prefix() + "-n"

class SchemaNodeException(YangsonException):
    """Abstract exception class for schema node errors."""

    def __init__(self, sn: SchemaNode):
        self.schema_node = sn

    def __str__(self) -> str:
        return str(self.schema_node.qual_name)

class NonexistentSchemaNode(SchemaNodeException):
    """A schema node doesn't exist."""

    def __init__(self, sn: SchemaNode, name: YangIdentifier,
                 ns: YangIdentifier = None):
        super().__init__(sn)
        self.name = name
        self.ns = ns

    def __str__(self) -> str:
        return "under {} – name '{}', namespace '{}'".format(
            super().__str__(), self.name, self.ns)

class BadSchemaNodeType(SchemaNodeException):
    """A schema node is of a wrong type."""

    def __init__(self, sn: SchemaNode, expected: str):
        super().__init__(sn)
        self.expected = expected

    def __str__(self) -> str:
        return super().__str__() + " is not a " + self.expected

class BadLeafrefPath(SchemaNodeException):
    """A leafref path is incorrect."""
    pass

class ValidationError(YangsonException):
    """Abstract exception class for instance validation errors."""

    def __init__(self, inst: "InstanceNode", detail: str):
        self.inst = inst
        self.detail = detail

    def __str__(self) -> str:
        return "[{}] {}".format(self.inst.json_pointer(), self.detail)

class SchemaError(ValidationError):
    """An instance violates a schema constraint."""
    pass

class SemanticError(ValidationError):
    """An instance violates a semantic rule."""
    pass

from .xpathast import Expr, LocationPath, Step, Root
from .instance import (InstanceNode, ArrayEntry,
                       NonexistentInstance, ObjectMember)



# Copyright © 2016 CZ.NIC, z. s. p. o.
#
# This file is part of Yangson.
#
# Yangson is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Yangson is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with Yangson.  If not, see <http://www.gnu.org/licenses/>.

"""Classes representing YANG data types.

This module implements the following classes:

* BitsType: YANG bits type.
* BinaryType: YANG binary type.
* BooleanType: YANG boolean type.
* DataType: Abstract class for data types.
* Decimal64Type: YANG decimal64 type.
* EmptyType: YANG empty type.
* EnumerationType: YANG enumeration type.
* IdentityrefType: YANG identityref type.
* InstanceIdentifierType: YANG instance-identifier type.
* LeafrefType: YANG leafref type.
* LinkType: Abstract class for data types representing links.
* IntegralType: Abstract class for integral types.
* Int8Type: YANG int8 type.
* Int16Type: YANG int16 type.
* Int32Type: YANG int32 type.
* Int64Type: YANG int64 type.
* NumericType: Abstract class for numeric types.
* StringType: YANG string type.
* Uint8Type: YANG uint8 type.
* Uint16Type: YANG uint16 type.
* Uint32Type: YANG uint32 type.
* Uint64Type: YANG uint64 type.
* UnionType: YANG union type.

The module also defines the following exceptions:

* YangTypeError: A scalar value is of incorrect type.
"""

import base64
import decimal
import re
from pyxb.utils.xmlre import XMLToPython
from typing import Any, Callable, List, Optional, Tuple, Union
from .exceptions import YangsonException
from .context import Context
from .instance import InstanceNode, InstanceIdParser, InstanceRoute
from .parser import ParserException
from .statement import Statement
from .typealiases import *
from .typealiases import _Singleton
from .xpathparser import XPathParser

# Local type aliases
Range = List[List[Union[int, decimal.Decimal]]]

class DataType:
    """Abstract class for YANG data types."""

    def __init__(self, mid: ModuleId):
        """Initialize the class instance."""
        self.module_id = mid
        self.default = None

    def from_raw(self, raw: RawScalar) -> ScalarValue:
        """Return a cooked value of the receiver type.

        Args:
            raw: Raw value obtained from JSON parser.

        Raises:
            YangTypeError: If `raw` doesn't conform to the receiver.
        """
        try:
            res = self._convert_raw(raw)
            if res is not None and self._constraints(res): return res
        except TypeError:
            raise YangTypeError(raw) from None
        raise YangTypeError(raw)

    def to_raw(self, val: ScalarValue) -> RawScalar:
        """Return a raw value ready to be serialized in JSON."""
        return val

    def parse_value(self, text: str) -> ScalarValue:
        """Parse value of a data type.

        Args:
            text: String representation of the value.
        """
        res = self._parse(text)
        if res is not None and self._constraints(res): return res
        raise YangTypeError(text)

    def canonical_string(self, val: ScalarValue) -> str:
        """Return canonical form of a value."""
        return str(val)

    def from_yang(self, text: str, mid: ModuleId) -> ScalarValue:
        """Parse value specified in a YANG module."""
        return self.parse_value(text)

    def contains(self, val: ScalarValue) -> bool:
        """Return ``True`` if the receiver type contains `val`."""
        try:
            return self._constraints(val)
        except TypeError:
            return False

    def _convert_raw(self, raw: RawScalar) -> ScalarValue:
        """Return a cooked value."""
        return raw

    def _constraints(self, val: Any) -> bool:
        return True

    @classmethod
    def _resolve_type(cls, stmt: Statement, mid: ModuleId) -> "DataType":
        typ = stmt.argument
        if typ in cls.dtypes:
            res = cls.dtypes[typ](mid)
            res._handle_properties(stmt, mid)
        else:
            res = cls._derived_type(stmt, mid)
        return res

    @classmethod
    def _derived_type(cls, stmt: Statement, mid: ModuleId) -> "DataType":
        """Completely resolve a derived type.

        Args:
            stmt: Derived type statement.
            mid: Id of the context module.
        """
        tchain = []
        s = stmt
        m = mid
        while True:
            tdef, m = Context.get_definition(s, m)
            s = tdef.find1("type", required=True)
            tchain.append((tdef, s, m))
            if s.argument in cls.dtypes: break
        res = cls.dtypes[s.argument](mid)
        btyp = True
        while tchain:
            tdef, typst, tid = tchain.pop()
            if btyp:
                res._handle_properties(s, m)
                btyp = False
            else:
                res._handle_restrictions(typst, tid)
            dfst = tdef.find1("default")
            if dfst:
                res.default = res.from_yang(dfst.argument, mid)
        res._handle_restrictions(stmt, mid)
        return res

    @staticmethod
    def _in_range(num: Union[int, decimal.Decimal], rng: Range) -> bool:
        """Decide whether a number fits into a range.

        Args:
            num: A number.
            rng: Numeric range.
        """
        for r in rng:
            if len(r) == 1:
                if r[0] == num: return True
            elif r[0] <= num <= r[1]: return True
        return False

    @staticmethod
    def _combine_ranges(orig: Range, rex: str,
                       parser: Callable[[str], Any]) -> Range:
        """Combine original range with a new one specified in `rex`.

        Args:
            orig: Original range.
            rex: Range expression.
        """
        to_num = lambda xs: [ parser(x) for x in xs ]
        lo = orig[0][0]
        hi = orig[-1][-1]
        parts = [ p.strip() for p in rex.split("|") ]
        ran = [ [ i.strip() for i in p.split("..") ] for p in parts ]
        if ran[0][0] != "min":
            lo = parser(ran[0][0])
        if ran[-1][-1] != "max":
            hi = parser(ran[-1][-1])
        return (
            [[lo, hi]] if len(ran) == 1 else
            [[lo, parser(ran[0][-1])]] +
            [ to_num(r) for r in ran[1:-1] ] +
            [[parser(ran[-1][0]), hi]])

    def _parse(self, text: str) -> Optional[ScalarValue]:
        return self._convert_raw(text)

    def _deref(self, node: InstanceNode) -> List[InstanceNode]:
        return []

    def _handle_properties(self, stmt: Statement, mid: ModuleId) -> None:
        """Handle type substatements.

        Args:
            stmt: YANG ``type`` statement.
            mid: Id of the context module.
        """
        self._handle_restrictions(stmt, mid)

    def _handle_restrictions(self, stmt: Statement, mid: ModuleId) -> None:
        """Handle type restriction substatements.

        Args:
            stmt: YANG ``type`` statement.
            mid: Id of the context module.
        """
        pass

class EmptyType(DataType, metaclass=_Singleton):
    """Singleton class representing YANG "empty" type."""

    def canonical_string(self, val: Tuple[None]) -> str:
        return ""

    def _constraints(self, val: Tuple[None]) -> bool:
        return val == (None,)

    def _parse(self, text: str) -> Tuple[None]:
        return None if text else (None,)

    def _convert_raw(self, raw: List[None]) -> Tuple[None]:
        try:
            return tuple(raw)
        except TypeError:
            return None

class BitsType(DataType):
    """Class representing YANG "bits" type."""

    def __init__(self, mid: ModuleId):
        """Initialize the class instance."""
        super().__init__(mid)
        self.bit = {}

    def _convert_raw(self, raw: str) -> Tuple[str]:
        try:
            return tuple(raw.split())
        except AttributeError:
            return None

    def _constraints(self, val: Tuple[str]) -> bool:
        for b in val:
            if b not in self.bit: return False
        return True

    def to_raw(self, val: Tuple[str]) -> str:
        return self.canonical_string(val)

    def as_int(self, val: Tuple[str]) -> int:
        """Transform a "bits" value to an integer."""
        res = 0
        try:
            for b in val:
                res += 1 << self.bit[b]
        except KeyError:
            raise YangTypeError(val) from None
        return res

    def _handle_properties(self, stmt: Statement, mid: ModuleId) -> None:
        """Handle **bit** statements."""
        nextpos = 0
        for bst in stmt.find_all("bit"):
            if not Context.if_features(bst, mid):
                continue
            label = bst.argument
            pst = bst.find1("position")
            if pst:
                pos = int(pst.argument)
                self.bit[label] = pos
                if pos > nextpos:
                    nextpos = pos
            else:
                self.bit[label] = nextpos
            nextpos += 1

    def _handle_restrictions(self, stmt: Statement, mid: ModuleId) -> None:
        bst = stmt.find_all("bit")
        if not bst: return
        new = set([ b.argument for b in bst if Context.if_features(b, mid) ])
        for bit in set(self.bit) - new:
            del self.bit[bit]

    def canonical_string(self, val: Tuple[str]) -> str:
        try:
            items = [(self.bit[b], b) for b in val]
        except KeyError:
            raise YangTypeError(val) from None
        items.sort()
        return " ".join([x[1] for x in items])
        
class BooleanType(DataType):
    """Class representing YANG "boolean" type."""

    def contains(self, val: bool) -> bool:
        """Return ``True`` if the receiver type contains `val`.

        Args:
            val: Value to test.
        """
        return isinstance(val, bool)

    def _parse(self, text: str) -> bool:
        """Parse boolean value.

        Args:
            text: String representation of the value.
        """
        if text == "true": return True
        if text == "false": return False

    def canonical_string(self, val: bool) -> str:
        if val is True: return "true"
        if val is False: return "false"
        raise YangTypeError(val)

class StringType(DataType):
    """Class representing YANG "string" type."""

    length = [[0, 4294967295]] # type: Range

    def __init__(self, mid: ModuleId):
        """Initialize the class instance."""
        super().__init__(mid)
        self.patterns = [] # type: List[Pattern]
        self.invert_patterns = [] # type: List[Pattern]

    def _handle_restrictions(self, stmt: Statement, mid: ModuleId) -> None:
        """Handle type restrictions.

        Args:
            stmt: YANG string type statement.
            mid: Id of the context module.
        """
        lstmt = stmt.find1("length")
        if lstmt:
            self.length = self._combine_ranges(self.length,
                                                lstmt.argument, int)
        for pst in stmt.find_all("pattern"):
            pat = re.compile(XMLToPython(pst.argument))
            if pst.find1("modifier", "invert-match"):
                self.invert_patterns.append(pat)
            else:
                self.patterns.append(pat)

    def contains(self, val: str) -> bool:
        """Return ``True`` if the receiver type contains `val`.

        Args:
            val: Value to test.
        """
        return isinstance(val, str) and self._constraints(val)

    def _constraints(self, val: str) -> bool:
        if not self._in_range(len(val), self.length):
            return False
        for p in self.patterns:
            if not p.match(val): return False
        for p in self.invert_patterns:
            if p.match(val): return False
        return True

class BinaryType(StringType):
    """Class representing YANG "binary" type."""

    def _convert_raw(self, raw: str) -> bytes:
        try:
            return base64.b64decode(raw, validate=True)
        except TypeError:
            return None

    def contains(self, val: bytes) -> bool:
        return isinstance(val, bytes) and self._constraints(val)

    def _constraints(self, val: bytes) -> bool:
        return self._in_range(len(val), self.length)

    def to_raw(self, val: bytes) -> str:
        return self.canonical_string(val)

    def canonical_string(self, val: bytes) -> str:
        return base64.b64encode(val).decode("ascii")

class EnumerationType(DataType):
    """Class representing YANG "enumeration" type."""

    def __init__(self, mid: ModuleId):
        """Initialize the class instance."""
        super().__init__(mid)
        self.enum = {} # type: Dict[str, int]

    def _constraints(self, val: str) -> bool:
        return val in self.enum

    def _handle_properties(self, stmt: Statement, mid: ModuleId) -> None:
        """Handle **enum** statements."""
        nextval = 0
        for est in stmt.find_all("enum"):
            if not Context.if_features(est, mid):
                continue
            label = est.argument
            vst = est.find1("value")
            if vst:
                val = int(vst.argument)
                self.enum[label] = val
                if val > nextval:
                    nextval = val
            else:
                self.enum[label] = nextval
            nextval += 1

    def _handle_restrictions(self, stmt: Statement, mid: ModuleId) -> None:
        est = stmt.find_all("enum")
        if not est: return
        new = set([ e.argument for e in est if Context.if_features(e, mid) ])
        for en in set(self.enum) - new:
            del self.enum[en]

class LinkType(DataType):
    """Abstract class for instance-referencing types."""

    def __init__(self, mid: ModuleId):
        """Initialize the class instance."""
        super().__init__(mid)
        self.require_instance = True # type: bool

    def _handle_restrictions(self, stmt: Statement, mid: ModuleId) -> None:
        """Handle type substatements.

        Args:
            stmt: YANG ``type leafref/instance-identifier`` statement.
            mid: Id of the context module.
        """
        if stmt.find1("require-instance", "false"):
            self.require_instance = False

class LeafrefType(LinkType):
    """Class representing YANG "leafref" type."""

    def __init__(self, mid: ModuleId):
        """Initialize the class instance."""
        super().__init__(mid)
        self.path = None
        self.ref_type = None

    def _handle_properties(self, stmt: Statement, mid: ModuleId) -> None:
        """Handle type substatements.

        Args:
            stmt: YANG ``type leafref`` statement.
            mid: Id of the context module.
        """
        self.path = XPathParser(
            stmt.find1("path", required=True).argument, mid).parse()

    def canonical_string(self, val: ScalarValue) -> str:
        return self.ref_type.canonical_string(val)

    def _constraints(self, val: ScalarValue) -> bool:
        return self.ref_type._constraints(val)

    def _convert_raw(self, raw: RawScalar) -> ScalarValue:
        return self.ref_type._convert_raw(raw)

    def to_raw(self, val: ScalarValue) -> RawScalar:
        return self.ref_type.to_raw(val)

    def _deref(self, node: InstanceNode) -> List[InstanceNode]:
        ns = self.path.evaluate(node)
        return [n for n in ns if str(n) == str(node)]

class InstanceIdentifierType(LinkType):
    """Class representing YANG "instance-identifier" type."""

    def _convert_raw(self, raw: str) -> InstanceRoute:
        try:
            return InstanceIdParser(raw).parse()
        except ParserException:
            raise YangTypeError(raw) from None

    def to_raw(self, val: InstanceRoute) -> str:
        """Override the superclass method."""
        return str(val)

    def _deref(self, node: InstanceNode) -> List[InstanceNode]:
        return [node.top().goto(node.value)]

class IdentityrefType(DataType):
    """Class representing YANG "identityref" type."""

    def __init__(self, mid: ModuleId):
        """Initialize the class instance."""
        super().__init__(mid)
        self.bases = [] # type: List[QualName]

    def _convert_raw(self, raw: str) -> QualName:
        try:
            i1, s, i2 = raw.partition(":")
        except AttributeError:
            return None
        return (i2, i1) if s else (i1, Context.namespace(self.module_id))

    def _constraints(self, val: QualName) -> bool:
        for b in self.bases:
            if not Context.is_derived_from(val, b): return False
        return True

    def to_raw(self, val: QualName) -> str:
        return self.canonical_string(val)

    def from_yang(self, text: str, mid: ModuleId) -> QualName:
        """Override the superclass method."""
        try:
            res = Context.translate_pname(text, mid)
        except:
            raise YangTypeError(text) from None
        if self._constraints(res): return res
        raise YangTypeError(text)

    def _handle_properties(self, stmt: Statement, mid: ModuleId) -> None:
        """Handle type substatements.

        Args:
            stmt: YANG ``type identityref`` statement.
            mid: Id of the context module.
        """
        self.bases = [ Context.translate_pname(b.argument, mid)
                       for b in stmt.find_all("base") ]

    def canonical_string(self, val: ScalarValue) -> str:
        """Return canonical form of a value."""
        return "{}:{}".format(val[1], val[0])

class NumericType(DataType):
    """Abstract class for numeric data types."""

    def _constraints(self, val: Union[int, decimal.Decimal]) -> bool:
        return self._in_range(val, self._range)

    def _handle_restrictions(self, stmt: Statement, mid: ModuleId) -> None:
        """Handle type substatements.

        Args:
            stmt: YANG ``type`` statement.
            mid: Id of the context module.
        """
        rstmt = stmt.find1("range")
        if rstmt:
            self._range = self._combine_ranges(self._range, rstmt.argument,
                                               self.parse_value)

class Decimal64Type(NumericType):
    """Class representing YANG "decimal64" type."""

    def __init__(self, mid: ModuleId):
        """Initialize the class instance."""
        super().__init__(mid)
        self._epsilon = decimal.Decimal(0) # type: decimal.Decimal

    def _handle_properties(self, stmt: Statement, mid: ModuleId) -> None:
        """Handle type substatements.

        Args:
            stmt: YANG ``type decimal64`` statement.
            mid: Id of the context module.
        """
        fd = int(stmt.find1("fraction-digits", required=True).argument)
        self._epsilon = decimal.Decimal(10) ** -fd
        quot = decimal.Decimal(10**fd)
        lim = decimal.Decimal(9223372036854775808)
        self._range = [[-lim / quot, (lim - 1) / quot]]
        super()._handle_properties(stmt, mid)

    def _convert_raw(self, raw: str) -> decimal.Decimal:
        try:
            return decimal.Decimal(raw).quantize(self._epsilon)
        except decimal.InvalidOperation:
            return None

    def to_raw(self, val: decimal.Decimal) -> str:
        return self.canonical_string(val)

    def canonical_string(self, val: decimal.Decimal) -> str:
        return "0.0" if val == 0 else str(val).rstrip("0")

    def contains(self, val: decimal.Decimal) -> bool:
        """Return ``True`` if the receiver type contains `val`.

        Args:
            val: Value to test.
        """
        return isinstance(val, decimal.Decimal) and self._constraints(val)

class IntegralType(NumericType):
    """Abstract class for integral data types."""

    # Regular expressions
    hexa_re = re.compile(r"\s*(\+|-)?0x")

    def _parse(self, text: str) -> int:
        """Parse integral value.

        Args:
            text: String representation of the value.
        """
        try:
            return (int(text, 16) if self.hexa_re.match(text) else int(text))
        except ValueError:
            return None

    def _convert_raw(self, raw: str) -> int:
        try:
            return int(raw)
        except (ValueError, TypeError):
            return None

    def contains(self, val: int) -> bool:
        """Return ``True`` if the receiver type contains `val`.

        Args:
            val: Value to test.
        """
        return isinstance(val, int) and self._constraints(val)

class Int8Type(IntegralType):
    """Class representing YANG "int8" type."""

    _range = [[-128,127]] # type: Range

class Int16Type(IntegralType):
    """Class representing YANG "int16" type."""

    _range = [[-32768, 32767]] # type: Range

class Int32Type(IntegralType):
    """Class representing YANG "int32" type."""

    _range = [[-2147483648, 2147483647]]  # type: Range

class Int64Type(IntegralType):
    """Class representing YANG "int64" type."""

    _range = [[-9223372036854775808, 9223372036854775807]] # type: Range

    def to_raw(self, val: int) -> str:
        return self.canonical_string(val)

class Uint8Type(IntegralType):
    """Class representing YANG "uint8" type."""

    _range = [[0, 255]] # type: Range

class Uint16Type(IntegralType):
    """Class representing YANG "uint16" type."""

    _range = [[0, 65535]] # type: Range

class Uint32Type(IntegralType):
    """Class representing YANG "uint32" type."""

    _range = [[0, 4294967295]] # type: Range

class Uint64Type(IntegralType):
    """Class representing YANG "uint64" type."""

    _range = [[0, 18446744073709551615]] # type: Range

    def _convert_raw(self, raw: str) -> int:
        try:
            return int(raw)
        except ValueError:
            return None

    def to_raw(self, val: int) -> str:
        return self.canonical_string(val)

class UnionType(DataType):
    """Class representing YANG "union" type."""

    def __init__(self, mid: ModuleId):
        """Initialize the class instance."""
        super().__init__(mid)
        self.types = [] # type: List[DataType]

    def to_raw(self, val: ScalarValue) -> RawScalar:
        for t in self.types:
            if t.contains(val):
                return t.to_raw(val)

    def canonical_string(self, val: ScalarValue) -> str:
        for t in self.types:
            if t.contains(val):
                return t.canonical_string(val)

    def _parse(self, text: str) -> Optional[ScalarValue]:
        for t in self.types:
            val = t._parse(text)
            if val is not None and t._constraints(val): return val
        return None

    def _convert_raw(self, raw: RawScalar) -> Optional[ScalarValue]:
        for t in self.types:
            val = t._convert_raw(raw)
            if val is not None and t._constraints(val): return val
        return None

    def _constraints(self, val: Any) -> bool:
        for t in self.types:
            try:
                if t._constraints(val): return True
            except TypeError:
                continue
        return False

    def _handle_properties(self, stmt: Statement, mid: ModuleId) -> None:
        """Handle type substatements.

        Args:
            stmt: YANG ``type`` statement.
            mid: Id of the context module.
        """
        self.types = [ self._resolve_type(ts, mid)
                       for ts in stmt.find_all("type") ]

class YangTypeError(YangsonException):
    """A value doesn't match its type."""

    def __init__(self, value):
        self.value = value

    def __str__(self) -> str:
        return "value " + repr(self.value)

DataType.dtypes = { "binary": BinaryType,
                    "bits": BitsType,
                    "boolean": BooleanType,
                    "decimal64": Decimal64Type,
                    "empty": EmptyType,
                    "enumeration": EnumerationType,
                    "identityref": IdentityrefType,
                    "instance-identifier": InstanceIdentifierType,
                    "int8": Int8Type,
                    "int16": Int16Type,
                    "int32": Int32Type,
                    "int64": Int64Type,
                    "leafref": LeafrefType,
                    "string": StringType,
                    "uint8": Uint8Type,
                    "uint16": Uint16Type,
                    "uint32": Uint32Type,
                    "uint64": Uint64Type,
                    "union": UnionType
                    }
"""Dictionary mapping type names to classes."""



from setuptools import setup
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))
with codecs.open(os.path.join(here, 'README.rst'), encoding='utf-8') as readme:
    long_description = readme.read()

setup(
    name = "yangson",
    packages = ["yangson"],
    version = "1.0.0rc3",
    description = "Library for working with data modelled in YANG",
    long_description = long_description,
    url = "https://github.com/CZ-NIC/yangson",
    author = "Ladislav Lhotka",
    author_email = "lhotka@nic.cz",
    license = "LGPLv3",
    install_requires = ['PyXB'],
    tests_require = ["pytest"],
    keywords = ["yang", "data model", "configuration", "json"],
    classifiers = [
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.5",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries",
        "Topic :: System :: Systems Administration"]
    )



# Copyright © 2016 CZ.NIC, z. s. p. o.
#
# This file is part of Yangson.
#
# Yangson is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Yangson is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with Yangson.  If not, see <http://www.gnu.org/licenses/>.

"""Instance data represented as a persistent structure.

This module implements the following classes:

* InstanceNode: Abstract class for instance nodes.
* RootNode: Root of the data tree.
* ObjectMember: Instance node that is an object member.
* ArrayEntry: Instance node that is an array entry.
* InstanceRoute: Route into an instance value.
* ResourceIdParser: Parser for RESTCONF resource identifiers.
* InstanceIdParser: Parser for instance identifiers.

The module defines the following exceptions:

* InstanceException: Base class for exceptions related to operations
  on instance nodes.
* InstanceValueError: The instance value is incompatible with the called method.
* NonexistentInstance: Attempt to access an instance node that doesn't
  exist.
"""

from datetime import datetime
from typing import Any, Callable, List, Tuple
from urllib.parse import unquote
from .exceptions import YangsonException
from .context import Context
from .enumerations import ContentType
from .instvalue import *
from .parser import EndOfInput, Parser, UnexpectedInput
from .typealiases import *

class InstanceNode:
    """YANG data node instance implemented as a zipper structure."""

    def __init__(self, value: Value, parinst: Optional["InstanceNode"],
                 schema_node: "DataNode", timestamp: datetime):
        """Initialize the class instance."""
        self.parinst = parinst         # type: Optional["InstanceNode"]
        """Parent instance node, or ``None`` for the root node."""
        self.schema_node = schema_node # type: DataNode
        """Data node corresponding to the instance node."""
        self.timestamp = timestamp     # type: datetime
        """Time of the receiver's last modification."""
        self.value = value             # type: Value
        """Value of the receiver."""

    @property
    def namespace(self) -> Optional[YangIdentifier]:
        """The receiver's namespace."""
        return self.schema_node.ns

    @property
    def qual_name(self) -> Optional[QualName]:
        """The receiver's qualified name."""
        return None

    def __str__(self) -> str:
        """Return string representation of the receiver's value."""
        sn = self.schema_node
        return (str(self.value) if isinstance(self.value, StructuredValue) else
                sn.type.canonical_string(self.value))

    def is_internal(self) -> bool:
        """Return ``True`` if the receiver is an instance of an internal node."""
        return isinstance(self.schema_node, InternalNode)

    def json_pointer(self) -> str:
        """Return JSON Pointer [RFC6901]_ of the receiver."""
        parents = []
        inst = self
        while inst.parinst:
            parents.append(inst)
            inst = inst.parinst
        return "/" + "/".join([i._pointer_fragment() for i in parents[::-1]])

    def member(self, name: InstanceName) -> "ObjectMember":
        """Return an instance node corresponding to a receiver's member.

        Args:
            name: Instance name of the member

        Raises:
            NonexistentSchemaNode: If the member isn't permitted by the schema.
            NonexistentInstance: If receiver's value doesn't contain member
                `name`.
            InstanceValueError: If the receiver's value is not an object.
        """
        if not isinstance(self.value, ObjectValue):
            raise InstanceValueError(self, "member of non-object")
        csn = self._member_schema_node(name)
        sibs = self.value.copy()
        try:
            return ObjectMember(name, sibs, sibs.pop(name), self,
                                csn, self.value.timestamp)
        except KeyError:
            raise NonexistentInstance(self, "member " + name) from None

    def put_member(self, name: InstanceName, value: Value,
                   raw: bool = False) -> "InstanceNode":
        """Return receiver's member with a new value.

        If the member is permitted by the schema but doesn't exist, it
        is created.

        Args:
            name: Instance name of the member.
            value: New value of the member.
            raw: Flag to be set if `value` is raw.

        Raises:
            NonexistentSchemaNode: If member `name` is not permitted by the
                schema.
            InstanceValueError: If the receiver's value is not an object.
        """
        if not isinstance(self.value, ObjectValue):
            raise InstanceValueError(self, "member of non-object")
        csn = self._member_schema_node(name)
        newval = self.value.copy()
        newval[name] = csn.from_raw(value) if raw else value
        ts = datetime.now()
        return self._copy(ObjectValue(newval, ts) , ts).member(name)

    def delete_member(self, name: InstanceName) -> "InstanceNode":
        """Return a copy of the receiver with a member deleted from its value.

        Args:
            name: Instance name of the member.

        Raises:
            NonexistentInstance: If receiver's value doesn't contain member
                `name`.
            InstanceValueError: If the receiver's value is not an object.
        """
        if not isinstance(self.value, ObjectValue):
            raise InstanceValueError(self, "member of non-object")
        newval = self.value.copy()
        try:
            del newval[name]
        except KeyError:
            raise NonexistentInstance(self, "member " + name) from None
        ts = datetime.now()
        return self._copy(ObjectValue(newval, ts), ts)

    def look_up(self, keys: Dict[InstanceName, ScalarValue]) -> "ArrayEntry":
        """Return the entry with matching keys.

        Args:
            keys: Dictionary of list keys and their values.

        Raises:
            InstanceValueError: If the receiver's value is not a YANG list.
            NonexistentInstance: If no entry with matching keys exists.
        """
        if not isinstance(self.value, ArrayValue):
            raise InstanceValueError(self, "lookup on non-list")
        try:
            for i in range(len(self.value)):
                en = self.value[i]
                flag = True
                for k in keys:
                    if en[k] != keys[k]:
                        flag = False
                        break
                if flag: return self.entry(i)
            raise NonexistentInstance(self, "entry lookup failed")
        except KeyError:
            raise NonexistentInstance(self, "entry lookup failed") from None
        except TypeError:
            raise InstanceValueError(self, "lookup on non-list") from None

    def entry(self, index: int) -> "ArrayEntry":
        """Return an instance node corresponding to a receiver's entry.

        Args:
            index: Index of the entry.

        Raises:
            InstanceValueError: If the receiver's value is not an array.
            NonexistentInstance: If entry `index` is not present.
        """
        val = self.value
        if not isinstance(val, ArrayValue):
            raise InstanceValueError(self, "entry of non-array")
        try:
            return ArrayEntry(val[:index], val[index+1:], val[index], self,
                              self.schema_node, val.timestamp)
        except IndexError:
            raise NonexistentInstance(self, "entry " + str(index)) from None

    def last_entry(self) -> "ArrayEntry":
        """Return an instance node corresponding to the receiver's last entry.

        Raises:
            InstanceValueError: If the receiver's value is not an array.
            NonexistentInstance: If the receiver's value is an empty array.
        """
        val = self.value
        if not isinstance(val, ArrayValue):
            raise InstanceValueError(self, "last entry of non-array")
        try:
            return ArrayEntry(val[:-1], [], val[-1], self,
                              self.schema_node, val.timestamp)
        except IndexError:
            raise NonexistentInstance(self, "last of empty") from None

    def delete_entry(self, index: int) -> "InstanceNode":
        """Return a copy of the receiver with an entry deleted from its value.

        Args:
            index: Index of the deleted entry.

        Raises:
            InstanceValueError: If the receiver's value is not an array.
            NonexistentInstance: If entry `index` is not present.
        """
        val = self.value
        if not isinstance(val, ArrayValue):
            raise InstanceValueError(self, "entry of non-array")
        if index >= len(val):
            raise NonexistentInstance(self, "entry " + str(index)) from None
        ts = datetime.now()
        return self._copy(ArrayValue(val[:index] + val[index+1:], ts), ts)

    def up(self) -> "InstanceNode":
        """Return an instance node corresponding to the receiver's parent.

        Raises:
            NonexistentInstance: If there is no parent.
        """
        ts = max(self.timestamp, self.parinst.timestamp)
        return self.parinst._copy(self._zip(), ts)

    def top(self) -> "InstanceNode":
        """Return an instance node corresponding to the root of the data tree."""
        inst = self
        while inst.parinst:
            inst = inst.up()
        return inst

    def update(self, value: Union[RawValue, Value],
               raw: bool = False) -> "InstanceNode":
        """Update the receiver's value.

        Args:
            value: New value.
            raw: Flag to be set if `value` is raw.

        Returns:
            Copy of the receiver with the updated value.
        """
        newval = self.schema_node.from_raw(value) if raw else value
        return self._copy(newval, datetime.now())

    def goto(self, iroute: "InstanceRoute") -> "InstanceNode":
        """Move the focus to an instance inside the receiver's value.

        Args:
            iroute: Instance route (relative to the receiver).

        Returns:
            The instance node corresponding to the target instance.

        Raises:
            InstanceValueError: If `iroute` is incompatible with the receiver's
                value.
            NonexistentInstance: If the instance node doesn't exist.
        """
        inst = self
        for sel in iroute:
            inst = sel.goto_step(inst)
        return inst

    def peek(self, iroute: "InstanceRoute") -> Optional[Value]:
        """Return a value within the receiver's subtree.

        Args:
            iroute: Instance route (relative to the receiver).
        """
        val = self.value
        for sel in iroute:
            val = sel.peek_step(val)
            if val is None: return None
        return val

    def validate(self, ctype: ContentType = ContentType.config) -> None:
        """Perform schema validation of the receiver's value.

        Args:
            ctype: Receiver's content type.

        Raises:
            SchemaError: If the value doesn't conform to the schema.
            SemanticError: If the value violates a semantic constraint.
        """
        self.schema_node.validate(self, ctype)

    def raw_value(self) -> RawValue:
        """Return receiver's value in a raw form (ready for JSON encoding)."""
        if isinstance(self.value, ObjectValue):
            res = {}
            for m in self.value:
                res[m] = self.member(m).raw_value()
        elif isinstance(self.value, ArrayValue):
            res = []
            try:
                en = self.entry(0)
                while True:
                    res.append(en.raw_value())
                    en = en.next()
            except NonexistentInstance:
                pass
        else:
            res = self.schema_node.type.to_raw(self.value)
        return res

    def add_defaults(self, ctype: ContentType = None) -> "InstanceNode":
        """Return the receiver with defaults added recursively to its value.

        Args:
            ctype: Content type of the defaults to be added. If it is
                ``None``, the content type will be the same as receiver's.
        """
        sn = self.schema_node
        val = self.value
        if not (isinstance(val, ObjectValue) and isinstance(sn, InternalNode)):
            return self
        res = self
        if val:
            for mn in val:
                m = res.member(mn) if res is self else res.sibling(mn)
                res = m.add_defaults(ctype)
            res = res.up()
        return sn._add_defaults(res, ctype)

    def _peek_schema_route(self, sroute: SchemaRoute) -> Value:
        irt = InstanceRoute()
        sn = self.schema_node
        for qn in sroute:
            sn = sn.get_child(*qn)
            if sn is None:
                raise NonexistentSchemaNode(sn, *qn)
            if isinstance(sn, DataNode):
                irt.append(MemberName(sn.iname()))
        return self.peek(irt)

    def _member_schema_node(self, name: InstanceName) -> "DataNode":
        qname = self.schema_node._iname2qname(name)
        res = self.schema_node.get_data_child(*qname)
        if res is None:
            raise NonexistentSchemaNode(self.schema_node, *qname)
        return res

    def _node_set(self) -> List["InstanceNode"]:
        """XPath - return the list of all receiver's nodes."""
        val = self.value
        if isinstance(val, ArrayValue):
            return [ self.entry(i) for i in range(len(val)) ]
        return [self]

    def _children(self, qname:
                  Union[QualName, bool] = None) -> List["InstanceNode"]:
        """XPath - return the list of receiver's children."""
        sn = self.schema_node
        if not isinstance(sn, InternalNode): return []
        if qname:
            cn = sn.get_data_child(*qname)
            if cn is None: return []
            iname = cn.iname()
            if iname in self.value:
                return self.member(iname)._node_set()
            wd = cn._default_instance(self, ContentType.all, lazy=True)
            if iname not in wd.value: return []
            while True:
                cn = cn.parent
                if cn is sn:
                    return wd.member(iname)._node_set()
                if (cn.when and not cn.when.evaluate(self) or
                    isinstance(cn, CaseNode) and
                    cn.qual_name != cn.parent.default_case):
                    return []
        res = []
        wd = sn._add_defaults(self, ContentType.all, lazy=True)
        for mn in wd.value:
            res.extend(wd.member(mn)._node_set())
        return res

    def _descendants(self, qname: Union[QualName, bool] = None,
                    with_self: bool = False) -> List["InstanceNode"]:
        """XPath - return the list of receiver's descendants."""
        res = ([] if not with_self or (qname and self.qual_name != qname)
               else [self])
        for c in self._children():
            if not qname or c.qual_name == qname:
                res.append(c)
            res += c._descendants(qname)
        return res

    def _preceding_siblings(
            self, qname: Union[QualName, bool] = None) -> List["InstanceNode"]:
        """XPath - return the list of receiver's preceding-siblings."""
        return []

    def _following_siblings(
            self, qname: Union[QualName, bool] = None) -> List["InstanceNode"]:
        """XPath - return the list of receiver's following-siblings."""
        return []

    def _parent(self) -> List["InstanceNode"]:
        """XPath - return the receiver's parent as a singleton list."""
        return [self.up()]

    def _deref(self) -> List["InstanceNode"]:
        """XPath: return the list of nodes that the receiver refers to."""
        return ([] if self.is_internal() else
                self.schema_node.type._deref(self))

class RootNode(InstanceNode):
    """This class represents the root of the instance tree."""

    def __init__(self, value: Value, schema_node: "DataNode",
                 timestamp: datetime):
        super().__init__(value, None, schema_node, timestamp)
        self.name = None # type: None
        """The instance name of the root node is always ``None``."""

    def up(self) -> None:
        """Override the superclass method.

        Raises:
            NonexistentInstance: root node has no parent
        """
        raise NonexistentInstance(self, "up of top")

    def _copy(self, newval: Value = None,
              newts: datetime = None) -> InstanceNode:
        return RootNode(newval if newval else self.value, self.schema_node,
                          newts if newts else self._timestamp)

    def _ancestors_or_self(
            self, qname: Union[QualName, bool] = None) -> List["RootNode"]:
        """XPath - return the list of receiver's ancestors including itself."""
        return [self] if qname is None else []

    def _ancestors(
            self, qname: Union[QualName, bool] = None) -> List["RootNode"]:
        """XPath - return the list of receiver's ancestors."""
        return []

class ObjectMember(InstanceNode):
    """This class represents an object member."""

    def __init__(self, name: InstanceName, siblings: Dict[InstanceName, Value],
                 value: Value, parinst: InstanceNode,
                 schema_node: "DataNode", timestamp: datetime ):
        super().__init__(value, parinst, schema_node, timestamp)
        self.name = name # type: InstanceName
        """The instance name of the receiver."""
        self.siblings = siblings # type: Dict[InstanceName, Value]
        """Sibling members within the parent object."""

    @property
    def qual_name(self) -> Optional[QualName]:
        """Return the receiver's qualified name."""
        p, s, loc = self.name.partition(":")
        return (loc, p) if s else (p, self.namespace)

    def sibling(self, name: InstanceName) -> "ObjectMember":
        """Return an instance node corresponding to a sibling member.

        Args:
            name: Instance name of the sibling member.

        Raises:
            NonexistentSchemaNode: If member `name` is not permitted by the schema.
            NonexistentInstance: If sibling member `name` doesn't exist.
        """
        ssn = self.parinst._member_schema_node(name)
        try:
            sibs = self.siblings.copy()
            newval = sibs.pop(name)
            sibs[self.name] = self.value
            return ObjectMember(name, sibs, newval, self.parinst,
                                ssn, self.timestamp)
        except KeyError:
            raise NonexistentInstance(self, "member " + name) from None

    def _zip(self) -> ObjectValue:
        """Zip the receiver into an object and return it."""
        res = ObjectValue(self.siblings.copy(), self.timestamp)
        res[self.name] = self.value
        return res

    def _pointer_fragment(self) -> str:
        return self.name

    def _copy(self, newval: Value = None,
              newts: datetime = None) -> "ObjectMember":
        return ObjectMember(self.name, self.siblings,
                           self.value if newval is None else newval,
                           self.parinst, self.schema_node,
                           newts if newts else self._timestamp)

    def _ancestors_or_self(
            self, qname: Union[QualName, bool] = None) -> List[InstanceNode]:
        """XPath - return the list of receiver's ancestors including itself."""
        res = [] if qname and self.qual_name != qname else [self]
        return res + self.up()._ancestors_or_self(qname)

    def _ancestors(
            self, qname: Union[QualName, bool] = None) -> List[InstanceNode]:
        """XPath - return the list of receiver's ancestors."""
        return self.up()._ancestors_or_self(qname)

class ArrayEntry(InstanceNode):
    """This class represents an array entry."""

    def __init__(self, before: List[Value], after: List[Value],
                 value: Value, parinst: InstanceNode,
                 schema_node: "DataNode", timestamp: datetime = None):
        super().__init__(value, parinst, schema_node, timestamp)
        self.before = before # type: List[Value]
        """Preceding entries of the parent array."""
        self.after = after # type: List[Value]
        """Following entries of the parent array."""

    @property
    def index(self) -> int:
        """Return the receiver's index."""
        return len(self.before)

    @property
    def name(self) -> InstanceName:
        """Return the name of the receiver."""
        return self.parinst.name

    @property
    def qual_name(self) -> Optional[QualName]:
        """Return the receiver's qualified name."""
        return self.parinst.qual_name

    def update(self, value: Union[RawValue, Value],
               raw: bool = False) -> "ArrayEntry":
        """Update the receiver's value.

        This method overrides the superclass method.
        """
        return super().update(self._cook_value(value, raw), False)

    def previous(self) -> "ArrayEntry":
        """Return an instance node corresponding to the previous entry.

        Raises:
            NonexistentInstance: If the receiver is the first entry of the parent array.
        """
        try:
            newval = self.before[-1]
        except IndexError:
            raise NonexistentInstance(self, "previous of first") from None
        return ArrayEntry(self.before[:-1], [self.value] + self.after, newval,
                          self.parinst, self.schema_node, self.timestamp)

    def next(self) -> "ArrayEntry":
        """Return an instance node corresponding to the next entry.

        Raises:
            NonexistentInstance: If the receiver is the last entry of the parent array.
        """
        try:
            newval = self.after[0]
        except IndexError:
            raise NonexistentInstance(self, "next of last") from None
        return ArrayEntry(self.before + [self.value], self.after[1:], newval,
                          self.parinst, self.schema_node, self.timestamp)

    def insert_before(self, value: Union[RawValue, Value],
                      raw: bool = False) -> "ArrayEntry":
        """Insert a new entry before the receiver.

        Args:
            value: The value of the new entry.
            raw: Flag to be set if `value` is raw.

        Returns:
            An instance node of the new inserted entry.
        """
        return ArrayEntry(self.before, [self.value] + self.after,
                          self._cook_value(value, raw), self.parinst,
                          self.schema_node, datetime.now())

    def insert_after(self, value: Union[RawValue, Value],
                     raw: bool = False) -> "ArrayEntry":
        """Insert a new entry after the receiver.

        Args:
            value: The value of the new entry.
            raw: Flag to be set if `value` is raw.

        Returns:
            An instance node of the newly inserted entry.
        """
        return ArrayEntry(self.before + [self.value], self.after,
                          self._cook_value(value, raw), self.parinst,
                          self.schema_node, datetime.now())

    def _cook_value(self, value: Union[RawValue, Value], raw: bool) -> Value:
        return (super(SequenceNode, self.schema_node).from_raw(value) if raw
                else value)

    def _zip(self) -> ArrayValue:
        """Zip the receiver into an array and return it."""
        res = ArrayValue(self.before.copy(), self.timestamp)
        res.append(self.value)
        res += self.after
        return res

    def _pointer_fragment(self) -> int:
        return str(len(self.before))

    def _copy(self, newval: Value = None,
              newts: datetime = None) -> "ArrayEntry":
        return ArrayEntry(self.before, self.after,
                          newval if newval else self.value,
                          self.parinst, self.schema_node,
                          newts if newts else self._timestamp)

    def _ancestors_or_self(
            self, qname: Union[QualName, bool] = None) -> List[InstanceNode]:
        """XPath - return the list of receiver's ancestors including itself."""
        res = [] if qname and self.qual_name != qname else [self]
        return res + self.up()._ancestors(qname)

    def _ancestors(
            self, qname: Union[QualName, bool] = None) -> List[InstanceNode]:
        """XPath - return the list of receiver's ancestors."""
        return self.up()._ancestors(qname)

    def _preceding_siblings(
            self, qname: Union[QualName, bool] = None) -> List[InstanceNode]:
        """XPath - return the list of receiver's preceding siblings."""
        if qname and self.qual_name != qname:
            return []
        res = []
        ent = self
        for i in range(len(self.before)):
            ent = ent.previous()
            res.append(ent)
        return res

    def _following_siblings(
            self, qname: Union[QualName, bool] = None) -> List[InstanceNode]:
        """XPath - return the list of receiver's following siblings."""
        if qname and self.qual_name != qname:
            return []
        res = []
        ent = self
        for i in range(len(self.after)):
            ent = ent.next()
            res.append(ent)
        return res

    def _parent(self) -> List["InstanceNode"]:
        """XPath - return the receiver's parent as a singleton list."""
        return [self.up().up()]

class InstanceRoute(list):
    """This class represents a route into an instance value."""

    def __str__(self) -> str:
        """Return a string representation of the receiver."""
        return "".join([str(c) for c in self])

    def __hash__(self) -> int:
        """Return the hash value of the receiver."""
        return self.__str__().__hash__()

class InstanceSelector:
    """Components of instance identifers."""
    pass

class MemberName(InstanceSelector):
    """Selectors of object members."""

    def __init__(self, name: InstanceName):
        """Initialize the class instance.

        Args:
            name: Member name.
        """
        self.name = name

    def __str__(self) -> str:
        """Return a string representation of the receiver."""
        return "/" + self.name

    def __eq__(self, other: "MemberName") -> bool:
        return self.name == other.name

    def peek_step(self, obj: ObjectValue) -> Value:
        """Return the member of `obj` addressed by the receiver.

        Args:
            obj: Current object.
        """
        return obj.get(self.name)

    def goto_step(self, inst: InstanceNode) -> InstanceNode:
        """Return member instance of `inst` addressed by the receiver.

        Args:
            inst: Current instance.
        """
        return inst.member(self.name)

class EntryIndex(InstanceSelector):
    """Numeric selectors for a list or leaf-list entry."""

    def __init__(self, index: int):
        """Initialize the class instance.

        Args:
            index: Index of an entry.
        """
        self.index = index

    def __str__(self) -> str:
        """Return a string representation of the receiver."""
        return "[{0:d}]".format(self.index)

    def __eq__(self, other: "EntryIndex") -> bool:
        return self.index == other.index

    def peek_step(self, arr: ArrayValue) -> Value:
        """Return the entry of `arr` addressed by the receiver.

        Args:
            arr: Current array.
        """
        try:
            return arr[self.index]
        except IndexError:
            return None

    def goto_step(self, inst: InstanceNode) -> InstanceNode:
        """Return member instance of `inst` addressed by the receiver.

        Args:
            inst: Current instance.
        """
        return inst.entry(self.index)

class EntryValue(InstanceSelector):
    """Value-based selectors of an array entry."""

    def __init__(self, value: ScalarValue):
        """Initialize the class instance.

        Args:
            value: Value of a leaf-list entry.
        """
        self.value = value

    def __str__(self) -> str:
        """Return a string representation of the receiver."""
        return "[.=" + str(self.value) +"]"

    def __eq__(self, other: "EntryValue") -> bool:
        return self.value == other.value

    def peek_step(self, arr: ArrayValue) -> Value:
        """Return the entry of `arr` addressed by the receiver.

        Args:
            arr: Current array.
        """
        try:
            return arr[arr.index(self.value)]
        except ValueError:
            return None

    def goto_step(self, inst: InstanceNode) -> InstanceNode:
        """Return member instance of `inst` addressed by the receiver.

        Args:
            inst: Current instance.
        """
        try:
            return inst.entry(inst.value.index(self.value))
        except ValueError:
            raise NonexistentInstance(
                inst, "entry '{}'".format(str(self.value))) from None

class EntryKeys(InstanceSelector):
    """Key-based selectors for a list entry."""

    def __init__(self, keys: Dict[InstanceName, ScalarValue]):
        """Initialize the class instance.

        Args:
            keys: Dictionary with keys of an entry.
        """
        self.keys = keys

    def __str__(self) -> str:
        """Return a string representation of the receiver."""
        return "".join(["[{}={}]".format(k, repr(self.keys[k]))
                        for k in self.keys])

    def __eq__(self, other: "EntryKeys") -> bool:
        return self.keys == other.keys

    def peek_step(self, arr: ArrayValue) -> Value:
        """Return the entry of `arr` addressed by the receiver.

        Args:
            arr: Current array.
        """
        for en in arr:
            flag = True
            for k in self.keys:
                if en[k] != self.keys[k]:
                    flag = False
                    break
            if flag: return en

    def goto_step(self, inst: InstanceNode) -> InstanceNode:
        """Return member instance of `inst` addressed by the receiver.

        Args:
            inst: Current instance.
        """
        return inst.look_up(self.keys)

class InstancePathParser(Parser):
    """Abstract class for parsers of strings identifying instances."""

    def _member_name(self, sn: "InternalNode") -> Tuple[MemberName, "DataNode"]:
        """Parser object member name."""
        name, ns = self.prefixed_name()
        cn = sn.get_data_child(name, ns if ns else sn.ns)
        if cn is None:
            raise NonexistentSchemaNode(sn, name, ns)
        return (MemberName(cn.iname()), cn)

class ResourceIdParser(InstancePathParser):
    """Parser for RESTCONF resource identifiers."""

    def parse(self) -> InstanceRoute:
        """Parse resource identifier."""
        if self.peek() == "/": self.offset += 1
        res = InstanceRoute()
        sn = Context.schema
        while True:
            mnam, cn = self._member_name(sn)
            res.append(mnam)
            try:
                next = self.one_of("/=")
            except EndOfInput:
                return res
            if next == "=":
                res.append(self._key_values(cn))
                if self.at_end(): return res
            sn = cn

    def _key_values(self, sn: "SequenceNode") -> Union[EntryKeys, EntryValue]:
        """Parse leaf-list value or list keys."""
        try:
            keys = self.up_to("/")
        except EndOfInput:
            keys = self.remaining()
        if not keys:
            raise UnexpectedInput(self, "entry value or keys")
        if isinstance(sn, LeafListNode):
            return EntryValue(sn.type.parse_value(unquote(keys)))
        ks = keys.split(",")
        try:
            if len(ks) != len(sn.keys):
                raise UnexpectedInput(
                    self, "exactly {} keys".format(len(sn.keys)))
        except AttributeError:
            raise BadSchemaNodeType(sn, "list")
        sel = {}
        for j in range(len(ks)):
            knod = sn.get_data_child(*sn.keys[j])
            val = knod.type.parse_value(unquote(ks[j]))
            sel[knod.iname()] = val
        return EntryKeys(sel)

class InstanceIdParser(InstancePathParser):
    """Parser for YANG instance identifiers."""

    def parse(self) -> InstanceRoute:
        """Parse instance identifier."""
        res = InstanceRoute()
        sn = Context.schema
        while True:
            self.char("/")
            mnam, cn = self._member_name(sn)
            res.append(mnam)
            try:
                next = self.peek()
            except EndOfInput:
                return res
            if next == "[":
                self.offset += 1
                self.skip_ws()
                if self.peek() in "0123456789":
                    ind = self.unsigned_integer() - 1
                    if ind < 0:
                        raise UnexpectedInput(self, "positive index")
                    self.skip_ws()
                    self.char("]")
                    res.append(EntryIndex(ind))
                elif isinstance(cn, LeafListNode):
                    self.char(".")
                    res.append(EntryValue(self._get_value(cn)))
                else:
                    res.append(self._key_predicates(cn))
                if self.at_end(): return res
            sn = cn

    def _get_value(self, tn: "TerminalNode") -> ScalarValue:
        self.skip_ws()
        self.char("=")
        self.skip_ws()
        quote = self.one_of("'\"")
        val = self.up_to(quote)
        self.skip_ws()
        self.char("]")
        return tn.type.parse_value(val)

    def _key_predicates(self, sn: "ListNode") -> EntryKeys:
        "Parse one or more key predicates."""
        sel = {}
        while True:
            name, ns = self.prefixed_name()
            knod = sn.get_data_child(name, ns)
            val = self._get_value(knod)
            sel[knod.iname()] = val
            try:
                next = self.peek()
            except EndOfInput:
                break
            if next != "[": break
            self.offset += 1
            self.skip_ws()
        return EntryKeys(sel)

    def _key_values(self, sn: "SequenceNode") -> EntryKeys:
        """Parse leaf-list value or list keys."""
        try:
            keys = self.up_to("/")
        except EndOfInput:
            keys = self.remaining()
        if not keys:
            raise UnexpectedInput(self, "entry value or keys")
        if isinstance(sn, LeafListNode):
            return EntryValue(sn.type.parse_value(unquote(keys)))
        ks = keys.split(",")
        try:
            if len(ks) != len(sn.keys):
                raise UnexpectedInput(self,
                                      "exactly {} keys".format(len(sn.keys)))
        except AttributeError:
            raise BadSchemaNodeType(sn, "list")
        sel = {}
        for j in range(len(ks)):
            knod = sn.get_data_child(*sn.keys[j])
            val = knod.type.parse_value(unquote(ks[j]))
            sel[knod.iname()] = val
        return EntryKeys(sel)

class InstanceException(YangsonException):
    """Abstract class for exceptions related to operations on instance nodes."""

    def __init__(self, inst: InstanceNode):
        self.instance = inst

    def __str__(self):
        return "[" + self.instance.json_pointer() + "]"

class InstanceValueError(InstanceException):
    """The instance value is incompatible with the called method."""

    def __init__(self, inst: InstanceNode, detail: str):
        super().__init__(inst)
        self.detail = detail

    def __str__(self):
        return "{} {}".format(super().__str__(), self.detail)

class NonexistentInstance(InstanceException):
    """Attempt to access an instance node that doesn't exist."""

    def __init__(self, inst: InstanceNode, detail: str):
        super().__init__(inst)
        self.detail = detail

    def __str__(self):
        return "{} {}".format(super().__str__(), self.detail)

from .schema import (AnydataNode, CaseNode, ChoiceNode, DataNode, InternalNode,
                     LeafNode, LeafListNode, ListNode,
                     NonexistentSchemaNode, SequenceNode, TerminalNode)



# Copyright © 2016 CZ.NIC, z. s. p. o.
#
# This file is part of Yangson.
#
# Yangson is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Yangson is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with Yangson.  If not, see <http://www.gnu.org/licenses/>.

"""Type aliases for use with type hints [PEP484]_."""

from decimal import Decimal
from typing import Any, Dict, List, Optional, Tuple, Union

RevisionDate = str
"""RevisionDate in the format ``YYYY-MM-DD``, or empty string."""

YangIdentifier = str
"""YANG identifier, see sec. `6.2`_ of [RFC7950]_."""

InstanceName = str
"""Object member name (simple or qualified), see sec. `4`_ of [RFC7951]_."""

PrefName = str
"""Name with optional prefix – [YangIdentifier ":"] YangIdentifier."""

InstanceIdentifier = str
"""YANG instance identifier, see sec. `6.11`_ of [RFC7951]_."""

ResourceIdentifier = str
"""RESTCONF resource identifier, see sec. `3.5.3`_ of [BBW16]_."""

ScalarValue = Union[int, Decimal, str, Tuple[None]]
"""Scalar value of an InstanceNode."""

QualName = Tuple[YangIdentifier, YangIdentifier]
"""Qualified name, tuple of name and module name."""

SchemaNodeId = str
"""Schema node identifier, see. sec. `6.5`_ in [RFC7950]_."""

SchemaRoute = List[QualName]
"""Schema route, a list of qualified names of schema nodes."""

SchemaPath = str
"""Schema path similar to instance identifier containing names of schema nodes."""

DataPath = str # same syntax as SchemaPath but containing only data nodes
"""SchemaPath containing only names of data nodes."""

ModuleId = Tuple[YangIdentifier, RevisionDate]
"""Module identifier: (YangIdentifier, RevisionDate)."""

RawScalar = Union[bool, int, str, List[None]]
"""Raw scalar value as produced by JSON parser."""

RawObject = Dict[InstanceName, "RawValue"]
"""Raw object as returned by JSON parser."""

RawEntry = Union[RawScalar, RawObject]
"""Raw entry of a leaf-list or list."""

RawList = List[RawObject]
"""List of raw objects."""

RawLeafList = List[RawScalar]
"""List of raw scalars."""

RawValue = Union[RawScalar, RawObject, RawList, RawLeafList]
"""Raw value of any type."""

class _Singleton(type):
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(_Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



# Copyright © 2016 CZ.NIC, z. s. p. o.
#
# This file is part of Yangson.
#
# Yangson is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Yangson is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with Yangson.  If not, see <http://www.gnu.org/licenses/>.

"""Basic access to the Yangson library.

This module implements the following class:

* DataModel: Basic entry point to the YANG data model.
"""

import hashlib
import json
from typing import Dict, List, Optional
from .exceptions import YangsonException
from .context import Context, BadYangLibraryData
from .instance import (EntryKeys, RootNode, MemberName,
                       InstanceIdParser, ResourceIdParser)
from .schema import DataNode, GroupNode, RawObject, SchemaNode
from .typealiases import *
from .typealiases import _Singleton

class DataModel(metaclass=_Singleton):
    """Basic user-level entry point to Yangson library.

    It is a singleton class, which means that only one instance can be
    created.
    """

    def __init__(self, yltxt: str, mod_path: List[str]):
        """Initialize the class instance.
        
        Args:
            yltxt: JSON text with YANG library data.
            mod_path: List of directories where to look for YANG modules.

        Raises:
            BadYangLibraryData: If YANG library data is invalid.
            FeaturePrerequisiteError: If a pre-requisite feature isn't
                supported.
            MultipleImplementedRevisions: If multiple revisions of an
                implemented module are listed in YANG library.
            ModuleNotFound: If a YANG module wasn't found in any of the
                directories specified in `mod_path`.
        """
        Context.schema = GroupNode()
        try:
            yl = json.loads(yltxt)
        except json.JSONDecodeError as e:
            raise BadYangLibraryData(str(e)) from None
        Context._from_yang_library(yl, mod_path)

    @classmethod
    def from_file(cls, name: str, mod_path: List[str] = ["."]) -> "DataModel":
        """Initialize the data model from a file with YANG library data.

        Args:
            name: Name of a file with YANG library data.
            mod_path: List of directories where to look for YANG modules.

        Returns:
            The data model instance.

        Raises:
            The same exceptions as the class constructor above.
        """
        with open(name, encoding="utf-8") as infile:
            yltxt = infile.read()
        return cls(yltxt, mod_path)

    @staticmethod
    def module_set_id() -> str:
        """Compute unique id of YANG modules comprising the data model.

        Returns:
            String consisting of hexadecimal digits.
        """
        fnames = sorted(["@".join(m) for m in Context.modules])
        return hashlib.sha1("".join(fnames).encode("ascii")).hexdigest()

    @staticmethod
    def from_raw(robj: RawObject) -> RootNode:
        """Create an instance node from a raw data tree.

        Args:
            robj: Dictionary representing a raw data tree.

        Returns:
            Root instance node.
        """
        cooked = Context.schema.from_raw(robj)
        return RootNode(cooked, Context.schema, cooked.timestamp)

    @staticmethod
    def get_schema_node(path: SchemaPath) -> Optional[SchemaNode]:
        """Return the schema node addressed by a schema path.

        Args:
            path: Schema path.

        Returns:
            Schema node if found in the schema, or ``None``.

        Raises:
            BadPath: If the schema path is invalid.
        """
        return Context.schema.get_schema_descendant(Context.path2route(path))

    @staticmethod
    def get_data_node(path: DataPath) -> Optional[DataNode]:
        """Return the data node addressed by a data path.

        Args:
            path: Data path.

        Returns:
            Data node if found in the schema, or ``None``.

        Raises:
            BadPath: If the schema path is invalid
        """
        addr = Context.path2route(path)
        node = Context.schema
        for p in addr:
            node = node.get_data_child(*p)
            if node is None: return None
        return node

    @staticmethod
    def ascii_tree() -> str:
        """Generate ASCII art representation of the schema tree.

        Returns:
            String with the ASCII tree.
        """
        return Context.schema._ascii_tree("")



# Copyright © 2016 CZ.NIC, z. s. p. o.
#
# This file is part of Yangson.
#
# Yangson is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Yangson is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with Yangson.  If not, see <http://www.gnu.org/licenses/>.

"""Common exception classes."""

from .typealiases import YangIdentifier

class YangsonException(Exception):
    """Base class for all Yangson exceptions."""
    pass
